/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#include "blitter.h"
#include "chunck.h"
#include <algorithm>
#include <cstring> // for memcpy

#ifdef DEBUG_CLIPPER
#include <iostream>
#endif

using namespace std;

blitter::blitter(char * ptr, const point & dstsize, const rect & src) throw() : 
			_ptr(ptr), 
			_clip(dstsize), 
			_src(src),
			_cur(src.left(), src.top()),
			_outside(false) {
#ifdef DEBUG_CLIPPER
	_clipped = 0;
	_clipped_block = 0;
	cout << "blitter::blitter(" << dstsize << ", " << src << ") called" << endl;
#endif
	assert(_ptr);
	assert(_clip.get_x() > 0 && _clip.get_y() > 0);
	assert(_src.width() > 0 && _src.height() > 0);
	assert(_src.left() < _clip.get_x() && _src.right() <= _clip.get_x());
	assert(_src.top() < _clip.get_y() && _src.bottom() <= _clip.get_y());
	_offset = _ptr + _clip.get_x() * _cur.get_y()  + _cur.get_x();
}

blitter::~blitter() {
#ifdef DEBUG_CLIPPER
	if(_clipped || _clipped_block) {
		cout << "blitter clipped " << _clipped << " pixels and " << _clipped_block << " blocks" << endl;
	}
#endif
}

void blitter::advance(int x, int y) throw() {
	if(y != 0) {
		//~ _offset += _src.left() + _clip.get_x() - _cur.get_x();
		_cur.set(_src.left() + x, _cur.get_y() + y);
	} else {
		_cur.get_x() += x;
		if(_cur.get_x() >= _src.right()) {
			_cur.set(_src.left(), _cur.get_y()+1);
			//~ _offset += _src.left() + _clip.get_x() - _cur.get_x() - x;
		} else {
			//~ _offset += x;
		}
	}
	_offset = _ptr + _clip.get_x() * _cur.get_y()  + _cur.get_x();
	_outside = ! _src.is_inside(_cur); 
}

void blitter::advance_block(int x, int y) throw() {
	//~ if(y != 0) {
		//~ _cur.set(_src.left() + x*4, _cur.get_y() + y*4);
	//~ } else {
		//~ _cur.get_x() += x*4;
		//~ if(_cur.get_x() >= _src.right()) {
			//~ _cur.set(_src.left(), _cur.get_y()+3);
		//~ }
	//~ }
	//~ _offset = _ptr + _clip.get_x() * _cur.get_y()  + _cur.get_x();
	//~ _outside = ! _src.is_inside(_cur); 
	advance(x*4, y*4);
	//~ cout << "blitter::advance_block(" << x << ", " << y << ") == " << _cur << ", " << _outside << ", " << static_cast<int>(_offset - _ptr) << endl;
}

void blitter::put(char data) throw(exception) {
	if(!_outside) {
		*_offset = data;
		advance();
	}
#ifdef DEBUG_CLIPPER
	else	_clipped ++;
#endif
}

void blitter::put(char data, unsigned int len) throw(exception) {
	while(len) {
		if(_outside) {
#ifdef DEBUG_CLIPPER
			_clipped += len;
#endif
			break;
		}
		int l = min(static_cast<int>(len), min(_clip.get_x() - _cur.get_x(), _src.right() - _cur.get_x()));
		len -= l;
		memset(_offset, data, l);
		advance(l);
	}
}

void blitter::blit(char * ptr, unsigned int len) throw(exception) {
	while(len) {
		if(_outside) {
#ifdef DEBUG_CLIPPER
			_clipped += len;
#endif
			break;
		}
		int l = min(static_cast<int>(len), min(_clip.get_x() - _cur.get_x(), _src.right() - _cur.get_x()));
		len -= l;
		memcpy(_offset, ptr, l);
		ptr += l;
		advance(l);
	}
}

void blitter::blit(chunck & src, unsigned int len) throw(exception) {
	while(len) {
		if(_outside) {
#ifdef DEBUG_CLIPPER
			_clipped += len;
#endif
			break;
		}
		int l = min(static_cast<int>(len), min(_clip.get_x() -_cur.get_x(), _src.right() - _cur.get_x()));
		len -= l;
		src.read(_offset, l);
		advance(l);
	}
}

void blitter::put_block(unsigned int data) throw(exception) {
	if(_cur.get_x() + 3 < _src.right() && _cur.get_y() + 3 < _src.bottom()) // This is clipping
	{
		assert((_clip.get_x() & 3) == 0);
		unsigned int * dst = reinterpret_cast<unsigned int *>(_offset);
		int line_size = _clip.get_x() >> 2;
		*dst = data; dst += line_size;
		*dst = data; dst += line_size;
		*dst = data; dst += line_size;
		*dst = data;
#ifdef DEBUG_CLIPPER
	} else {
		_clipped_block ++;
#endif				
	}
	advance_block();
}

void blitter::put_block(unsigned int d1, unsigned int d2, unsigned int d3, unsigned int d4) throw(exception) {
	if(_cur.get_x() + 3 < _src.right() && _cur.get_y() + 3 < _src.bottom()) // This is clipping
	{
		assert((_clip.get_x() & 3) == 0);
		unsigned int * dst = reinterpret_cast<unsigned int *>(_offset);
		int line_size = _clip.get_x() >> 2;
		*dst = d4; dst += line_size;
		*dst = d3; dst += line_size;
		*dst = d2; dst += line_size;
		*dst = d1;
#ifdef DEBUG_CLIPPER
	} else {
		_clipped_block ++;
#endif
	}
	advance_block();
}

void blitter::put_block(unsigned char * data) throw(exception) {
	if(_cur.get_x() + 3 < _src.right() && _cur.get_y() + 3 < _src.bottom()) // This is clipping
	{
		assert((_clip.get_x() & 3) == 0);
		unsigned int * dst = reinterpret_cast<unsigned int *>(_offset);
		int line_size = _clip.get_x() >> 2;
		unsigned int * src = reinterpret_cast<unsigned int *>(data);
		*dst = *src++; dst += line_size; 
		*dst = *src++; dst += line_size; 
		*dst = *src++; dst += line_size; 
		*dst = *src++;
#ifdef DEBUG_CLIPPER
	} else {
		_clipped_block ++;
#endif				
	}
	advance_block();
}

void blitter::put_block(chunck & src) throw(exception) {
	if(_cur.get_x() + 3 < _src.right() && _cur.get_y() + 3 < _src.bottom()) // This is clipping
	{
		assert((_clip.get_x() & 3) == 0);
		unsigned int * dst = reinterpret_cast<unsigned int *>(_offset);
		int line_size = _clip.get_x() >> 2;
		//~ unsigned int  src = src.get_dreinterpret_cast<unsigned int *>(data);
		*dst = src.get_dword(); dst += line_size; 
		*dst = src.get_dword(); dst += line_size; 
		*dst = src.get_dword(); dst += line_size; 
		*dst = src.get_dword();
#ifdef DEBUG_CLIPPER
	} else {
		_clipped_block ++;
#endif				
	}
	advance_block();
}

void blitter::block_copy(int offset) {
	if(_cur.get_x() + 3 < _src.right() && _cur.get_y() + 3 < _src.bottom()) // This is clipping
	{
		char  * dst = _offset;
		*reinterpret_cast<unsigned int *>(dst) = *reinterpret_cast<unsigned int *>(dst + offset);
		dst += _clip.get_x();
		*reinterpret_cast<unsigned int *>(dst) = *reinterpret_cast<unsigned int *>(dst + offset);
		dst += _clip.get_x();
		*reinterpret_cast<unsigned int *>(dst) = *reinterpret_cast<unsigned int *>(dst + offset);
		dst += _clip.get_x();
		*reinterpret_cast<unsigned int *>(dst) = *reinterpret_cast<unsigned int *>(dst + offset);
#ifdef DEBUG_CLIPPER
	} else {
		_clipped_block ++;
#endif
	}
	advance_block();
}
