/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __BLITTER_H_
#define __BLITTER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef DEBUG
# ifndef NO_DEBUG_CLIPPER
#  define DEBUG_CLIPPER
# endif
#else
# ifdef DEBUG_CLIPPER
#  error DEBUG_CLIPPER defined without DEBUG
# endif
#endif

#include <exception>
#include "rect.h"

class chunck;
/*! 	@brief class for handling blitting on a frame buffer

	This class allows to perform secure blitting to a frame buffer in several ways.
	This means that clipping is performed, so that only the part that you want to modify can be modified.
*/
class blitter {
private:
	char * _ptr;	//!< This is the pointer to the start of the frame buffer
	char * _offset;	//!< This is the current pointer in the frame buffer
	point _clip;		//!<  This is the size of the frame buffer (width/height)
	rect _src; 		//!< This is the size and position of the destination rectangle
	point _cur; 		//!< This is the current position in the destination rectangle
	bool _outside;	//!< flag that is set to \c true when the blitter reach the end of the destination rectangle
#ifdef DEBUG_CLIPPER
	int _clipped;
	int _clipped_block;
#endif
public:
	/*!	@brief constructor

		@param buffer the frame buffer to blit to
		@param dstsize the size of the frame buffer
		@param src the rectangle to blit to
	*/
	blitter(char * buffer, const point & dstsize, const rect & src) throw();
	virtual ~blitter();
	void blit(char *, unsigned int) throw(std::exception); //!< This method allows to blit directly some data from a buffer
	void blit(chunck &, unsigned int) throw(std::exception); //!< This method allows to blit directly some data from a chunck
	void put(char) throw(std::exception); //!< This method allows to blit one byte
	void put(char, unsigned int) throw(std::exception); //!< This method allows to blit one byte several times
	void advance(int = 1, int = 0) throw(); //!< This method allows to advance the current position in the blitter
	void advance_block(int = 1, int = 0) throw(); //!< This method allows to advance the current position in the blitter in terms of blocks
	void put_block(unsigned int) throw(std::exception); //!< This method allows to blit one block from an int value repeated 4 time
	void put_block(chunck &) throw(std::exception); //!< This method allows to blit one block directly read from a chunck
	void put_block(unsigned char *) throw(std::exception); //!< This method allows to blit one block directly from a buffer
	void put_block(unsigned int, unsigned int, unsigned int, unsigned int) throw(std::exception); //!< This method allows to blit one block from a 4 int value
	void block_copy(int); //!< This method allows to copy one block from another separated by the given offset
};

#endif
