/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __BRENDERER_H_
#define __BRENDERER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "renderer.h"
#include "palette.h"
#include "rect.h"

/*! 	@brief base class for easily creating ::renderer instances

	This class implements some function available in the ::renderer abstract class, so that 
	creation of subclasses of ::renderer is easier.
*/
class base_renderer : public renderer {
private:
	palette _pal;		//!< The current palette
	char * _data;		//!< The current frame buffer
	int _frame;			//!< The current frame number
	int _nbframes;		//!< The number of frames in the animation
	int _width;			//!< The current frame's width
	int _height;		//!< The current frame's height
	std::string _fname;	//!< The filename of the animation being played
protected:
	virtual void save(int frame = -1) = 0;

protected:
	const std::string & get_filename() const throw() { return _fname; };	//!< accessor for animation filename
	int get_nbframes() const throw() { return _nbframes; };	//!< accessor for number of frames
	int get_width() const throw() { return _width; };	//!< accessor for current width
	int get_height() const throw() { return _height; };	//!< accessor for current height
	const palette & pal() const throw() { return _pal; };	//!< accessor for current palette
	const char * data() const throw() { return _data; };	//!< accessor for current frame buffer
	void clean() throw();	//!< memory cleanup (deletes frame buffer)
	void set_frame(int f) throw() { _frame = f; };	//!< allows to change the frame number
public:
	int get_frame() const throw() { return _frame; };	//!< accessor for current frame number
	base_renderer() throw();
	virtual ~base_renderer() throw();
	
	virtual bool init_frame(const point & size) throw(std::exception);
	virtual char * lock_frame(int frame) throw(std::exception);
	virtual bool unlock_frame() throw();
	virtual bool flip_frame() throw();
	virtual bool set_palette(const palette & pal) throw();
	virtual bool start_decode(const std::string & fname, int version, int nbframes) throw() { _fname = fname; _nbframes = nbframes; return true; }
	virtual sound_renderer * get_sound_renderer() throw(std::exception) { return 0; };
};

/*! 	@brief A null ::renderer

	This class completely implements ::renderer, without actually doing anything.
	This class is useful for performance measurements.
*/
class null_renderer : public base_renderer {
protected:
	void save(int frame = -1) {};
public:
	null_renderer() throw() {};
	virtual ~null_renderer() throw() {};
	bool wait(int ms) throw() { return true; };
};

#endif
