/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#include <sstream>
#include <fstream>
#include <stdexcept>

#include "channel.h"
#include "endian.h"
#include "chunck.h"
#include "chunck_type.h"

#ifdef DEBUG_CHANNEL
#include <iostream>
#include <iomanip>
#endif


using namespace std;

void saud_channel::handleStrk(chunck & b) throw(exception) {
	int size = b.get_size();
	if(size != 14 && size != 10) {
		ostringstream oss;
		oss << "STRK has a invalid size : " << size;
		throw runtime_error(oss.str());
	}
#ifdef DEBUG_CHANNEL
	int unk1 = b.get_byte();
	int unk2 = b.get_byte();
	int unk3 = from_big_endian(b.get_dword());
	int unk4 = from_big_endian(b.get_dword());
	cout << "TRACK " << _track << " STRK(" << size << ") content : ";
	cout << unk1 << " " << unk2 << " " << unk3 << "  "<< unk4;
	if(size == 14) {
		int unk5 = from_big_endian(b.get_dword());
		cout << " " << unk5;
	}
	cout << endl;
#endif
}

void saud_channel::handleSmrk(chunck & b) throw(exception) {
#ifdef DEBUG_CHANNEL
	int size = b.get_size();
	if(size > 0) // throw runtime_error("SMRK has a invalid size");
	{
		cout << "TRACK " << _track << " SMRK(" << size << ") content : ";
		for(int i = 0; i < size; i++) {
			cout << setw(3) << static_cast<int>(b.get_byte()) << " ";
		}
		cout << endl;
	}
#endif
	_mark_reached = true;
}

void saud_channel::handleShdr(chunck & b) throw(exception) {
	int size = b.get_size();
	if(size != 4) throw runtime_error("SMRK has a invalid size");
#ifdef DEBUG_CHANNEL
	int unknown = b.get_dword();
	cout << "TRACK " << _track << " SHDR(" << size << ") content : " << unknown << endl;
#endif
}

bool saud_channel::processBuffer() throw(exception) {
#ifdef DEBUG_CHANNEL
	cout << "processBuffer(" << _track << ") : _cur_buffer_write_pos " << _cur_buffer_write_pos << ", _cur_buffer_read_pos " << _cur_buffer_read_pos << endl;
#endif
	while((_cur_buffer_write_pos - _cur_buffer_read_pos) > 8) {
		int available_size = _cur_buffer_write_pos - _cur_buffer_read_pos;
		chunck::type type = from_big_endian(*reinterpret_cast<unsigned int *>(_buffer + _cur_buffer_read_pos));
		unsigned int size = from_big_endian(*reinterpret_cast<unsigned int *>(_buffer + _cur_buffer_read_pos + 4));
#ifdef DEBUG_CHANNEL
		cout << "TRACK " << _track << ", chunck " << chunck::chunck_string(type) << ", " << available_size << "/" << size+8 << endl;
#endif
		cont_chunck * c = 0;
		switch(type) {
			case TYPE_STRK:
				assert(available_size >= (size + 8));
				c = new cont_chunck(reinterpret_cast<char*>(_buffer) + _cur_buffer_read_pos);
				handleStrk(*c);
				break;
			case TYPE_SMRK:
				assert(available_size >= (size + 8));
				c = new cont_chunck(reinterpret_cast<char*>(_buffer) + _cur_buffer_read_pos);
				handleSmrk(*c);
				break;
			case TYPE_SHDR:
				assert(available_size >= (size + 8));
				c = new cont_chunck(reinterpret_cast<char*>(_buffer) + _cur_buffer_read_pos);
				handleShdr(*c);
				break;
			case TYPE_SDAT: // Sound data !!!
				if(_data_start == -1) {
					_rdata_start = _data_start = _cur_buffer_read_pos + 8;
					_rdata_size = _data_size = size; 
				}
				break;
			default:
				ostringstream oss;
				oss << "unknown chunck in SAUD track : " << chunck::chunck_string(type) << endl;
				throw(runtime_error(oss.str()));
		}
		if(c) delete c;
		if(available_size >= (size + 8)) {
			_cur_buffer_read_pos += size + 8;
		} else break;
	}
	return true;
}

saud_channel::saud_channel(int track, int freq) throw() : 
			_track(track), 
			_nbframes(0), 
			_buffer(0), 
			_buffer_size(0), 
			_frequency(freq), 
			_cur_buffer_write_pos(0),
			_cur_buffer_read_pos(0),
			_data_start(-1), 
			_mark_reached(false) {
}

extern bool win32_save_as_wave(const string & fname, char * data, int size, int freq);

saud_channel::~saud_channel() throw() {
#ifdef DEBUG_CHANNEL
	cout << "~saud_channel(" << _track << ") size == " << _buffer_size << ", _cur_buffer_write_pos == " << _cur_buffer_write_pos << endl;
#endif
#if 0
	ostringstream oss;
	oss << "track" << _track << "_" << _flags << "_" << _volume << ".wav";
	win32_save_as_wave(oss.str(), reinterpret_cast<char *>(_buffer) + _rdata_start, _rdata_size, _frequency);
#endif
	if(_buffer) { 
		delete []_buffer; 
	} 
}

bool saud_channel::is_terminated() const throw() {
#ifdef DEBUG_CHANNEL
	cout << "saud_channel::is_terminated() : " << _cur_buffer_read_pos << ", " << _buffer_size << ", " << _data_start << ", " << _data_size << endl;
#endif
	if(_cur_buffer_read_pos < _buffer_size) return false; // Frame completely processed
	if(_data_start != -1 && _data_size == 0) return true;
	return false;
}

void saud_channel::recalcVolumeTable() throw(exception) {
	const int MAX_BALANCE = 100;
	int volume_left, volume_right;
	if(_balance < -MAX_BALANCE || _balance > MAX_BALANCE) {
		ostringstream oss;
		oss << "balance is out of range ! : " << _balance;
		throw runtime_error(oss.str());
	}
	int left_multiplier = MAX_BALANCE - _balance;
	int right_multiplier = MAX_BALANCE + _balance;
	volume_left = _volume * left_multiplier / (MAX_BALANCE*2);
	volume_right = _volume * right_multiplier / (MAX_BALANCE*2);
	if(volume_left < 0) volume_left = 0;
	if(volume_left > 128) volume_left = 128;
	if(volume_right < 0) volume_right = 0;
	if(volume_right > 128) volume_right = 128;
	for(int i = 0; i < 256; i++) {
		_voltable[0][i] = volume_left * static_cast<signed char>(i);
		_voltable[1][i] = volume_right * static_cast<signed char>(i);
	}
}

bool saud_channel::set_parameters(int nb, int flags, int volume, int balance) throw(exception) {
	if(_nbframes || _buffer) throw(runtime_error("channel initialized twice"));
	_nbframes = nb;
	_flags = flags; // bit 7 == IS_VOICE, bit 6 == IS_BACKGROUND_MUSIC, other ??
	_volume = volume;
	_balance = balance;
	_index = 0;
	//~ cerr << "channel::set_parameters() == " << _track << " : " << _volume << ", " << _balance << endl;
	recalcVolumeTable();
	//~ for(int i = 0; i < 256; i++)
		//~ _voltable[i] = vol1 * (signed char)i;
	return true;
}

bool saud_channel::check_parameters(int index, int nb, int flags, int volume, int balance) throw(exception) {
	if(++_index != index) throw runtime_error("invalid index in saud_channel::check_parameters()");
	if(_nbframes != nb) throw runtime_error("invalid duration in saud_channel::check_parameters()");
	if(_flags != flags) throw runtime_error("invalid flags in saud_channel::check_parameters()");
	if(_volume != volume || _balance != balance) {
		// throw runtime_error("invalid volume in saud_channel::check_parameters()");
		_volume = volume;
		_balance = balance;
		//~ cerr << "saud_channel::check_parameters() == " << _track << " : " << _volume << ", " << _balance << endl;
		recalcVolumeTable();
	}
	return true;
}

bool saud_channel::append_data(chunck & b, int size) throw(exception) {
	if(!_buffer) { // First call
		assert(size > 8);
		chunck::type saud_type = from_big_endian(b.get_dword());
		unsigned int saud_size = from_big_endian(b.get_dword());
		if(saud_type != TYPE_SAUD) throw(runtime_error("Invalid CHUNCK for saud_channel"));
		size -= 8;
		_buffer_size = saud_size;
		assert(_buffer_size);
		_buffer = new unsigned char[_buffer_size];
		if(!_buffer)  throw(runtime_error("saud_channel failed to allocate memory"));
		b.read(_buffer, size);
		_cur_buffer_write_pos = size;
	} else {
		assert(size <= (_buffer_size - _cur_buffer_write_pos));
		b.read(_buffer + _cur_buffer_write_pos, size);
		_cur_buffer_write_pos += size;
	}
	return processBuffer();
}

void saud_channel::read_sound_data(short * snd, int size) throw(std::exception) {
	if(_data_start == -1 || _data_size == 0) throw(runtime_error("invalid call to saud_channel::read_sound_data()"));
	
#ifdef DEBUG_CHANNEL
	int available =  min(_cur_buffer_write_pos - _data_start, _data_size);
	cout << "saud_channel::read_sound_data(" << _track << ") size == " << size << ", available == " << available << endl;
	assert(available >= size);
#endif
	for(int i = 0; i < size; i++) {
		snd[2*i] += _voltable[0][_buffer[_data_start+i] ^ 0x80];
		snd[2*i+1] += _voltable[1][_buffer[_data_start+i] ^ 0x80];
	}
	_data_start += size;
	_data_size   -= size;
}

int saud_channel::available_sound_data(void) const throw() {
	return min(_cur_buffer_write_pos - _data_start, _data_size); 
}

imuse_channel::imuse_channel(int track, int freq) throw() : 
			_track(track), 
			_buffer(0), 
			_buffer_size(0), 
			_frequency(freq), 
			_cur_buffer_write_pos(0),
			_cur_buffer_read_pos(0),
			_data_start(-1){
	_decoded_buffer = new short[_frequency*2];
	assert(_decoded_buffer);
	_resampled_buffer = new short[_frequency*2];
	assert(_resampled_buffer);
}

imuse_channel::~imuse_channel() throw() {
#ifdef DEBUG_CHANNEL
	cout << "~imuse_channel(" << _track << ") size == " << _buffer_size << ", _cur_buffer_write_pos == " << _cur_buffer_write_pos << endl;
#endif
#if 0
	ofstream o;
	ostringstream oss;
	oss << "track" << _track << "_" << _unk1 << "_" << _unk2 << ".dat";
	o.open(oss.str().c_str(), ios::out|ios::binary);
	o.write(reinterpret_cast<char*>(_buffer), _buffer_size);
	o.close();
#endif
	if(_buffer) { 
		delete []_buffer; 
	}
	delete []_resampled_buffer;
	delete []_decoded_buffer;
}

bool imuse_channel::is_terminated() const throw()
{
	//~ if(_size == 0) return true;
	//~ cerr << "imuse_channel::is_terminated(" << _track << ") _cur_buffer_read_pos == " << _cur_buffer_read_pos << ", _buffer_size == " << _buffer_size << ", _data_size == " << _data_size << endl;
	if(_cur_buffer_read_pos < (_buffer_size - 8)) return false; // Frame completely processed
	if(_data_start != -1 && available_sound_data() == 0) return true;
	return false;
}

bool imuse_channel::set_parameters(int nbframes, int size, int unk1, int unk2) throw(std::exception)
{
#ifdef DEBUG_CHANNEL
	cout << "imuse_channel::set_parameters(" << _track << ", " << nbframes<< ", " << size<< ", " << unk1<< ", " << unk2 << ")" << endl;
#endif	
	_nbframes = nbframes;
	_size = size;
	_unk1 = unk1;
	_unk2 = unk2;
	_index = 0;
	return true;
}

bool imuse_channel::check_parameters(int index, int nbframes, int size, int unk1, int unk2) throw(std::exception)
{
#ifdef DEBUG_CHANNEL
	cout << "imuse_channel::check_parameters(" << 
		index << ", " << 
		nbframes << ", " << 
		size << ", " << 
		unk1 << ", " << 
		unk2 << ") == " << 
		_track << ", " << 
		_index+1 << ", " << 
		_nbframes << ", " << 
		_size << ", " << 
		_unk1 << ", " << 
		_unk2 << endl;
	if(++_index != index) throw runtime_error("invalid index in imuse_channel::check_parameters()");
	if(_nbframes != nbframes) throw runtime_error("invalid nbframes in imuse_channel::check_parameters()");
	if(_size != size) throw runtime_error("invalid size in imuse_channel::check_parameters()");
	if(_unk1 != unk1) throw runtime_error("invalid unk1 in imuse_channel::check_parameters()");
	if(_unk2 != unk2) throw runtime_error("invalid unk2 in imuse_channel::check_parameters()");
	//~ if(_unk3 != unk3) throw runtime_error("invalid unk3 in imuse_channel::check_parameters()");
#endif	
	return true;
}

bool imuse_channel::append_data(chunck & b, int size) throw(std::exception)
{
#ifdef DEBUG_CHANNEL
	cout << "imuse_channel::append_data(" << _track << ") : " << size << " < " << _buffer_size<< endl;
#endif
	_size -= size;
	if(!_buffer) { // First call
		assert(size > 8);
		chunck::type imus_type = from_big_endian(b.get_dword());
		unsigned int imus_size = from_big_endian(b.get_dword());
		if(imus_type != TYPE_iMUS) throw runtime_error("Invalid CHUNCK for imuse_channel");
		size -= 8;
		_buffer_size = imus_size;
		assert(_buffer_size);
		_buffer = new unsigned char[_buffer_size];
		if(!_buffer)  throw runtime_error("imuse_channel failed to allocate memory");
		b.read(_buffer, size);
		_cur_buffer_write_pos = size;
	} else {
#ifdef DEBUG_CHANNEL
		if(size > (_buffer_size - _cur_buffer_write_pos))
			cerr << "imuse_channel::append_data(" << _track << ") : " << size << " > " << _buffer_size - _cur_buffer_write_pos << endl;
#endif
			//~ size = _buffer_size - _cur_buffer_write_pos;
		assert(size <= (_buffer_size - _cur_buffer_write_pos));
		
		b.read(_buffer + _cur_buffer_write_pos, size);
		_cur_buffer_write_pos += size;
	}
	return processBuffer();
}

bool imuse_channel::handleFormat(chunck & src) {
	if(src.get_size() != 20) throw runtime_error("invalid size for FRMT chunck");
	_imuse_start = from_big_endian(src.get_dword());
#ifdef DEBUG_CHANNEL
	int unknown = from_big_endian(src.get_dword());
#else
	src.seek(4);
#endif
	_bitsize = from_big_endian(src.get_dword());
	_rate = from_big_endian(src.get_dword());
	_channels = from_big_endian(src.get_dword());
	assert(_channels == 1 || _channels == 2);
#ifdef DEBUG_CHANNEL
	if(unknown != 0) cerr << "imuse_channel::handleFormat() == " << unknown << ", " << _imuse_start << ", " << _bitsize << ", " << _rate << ", " << _channels << endl;
#endif
	return true;
}

bool imuse_channel::handleText(chunck & src) {
	//~ if(src.get_size() != 20) throw runtime_error("invalid size for FRMT chunck");
#ifdef DEBUG_CHANNEL
	int start = from_big_endian(src.get_dword());
	assert(start == _imuse_start);
	int text_size = src.get_size() - 4;
	char * buf = new char[text_size];
	src.read(buf, text_size);
	string text(buf, text_size);
	delete []buf;
	assert(text.size() == text_size);
#endif
	//~ cerr << "imuse_channel::handleText() == " << text << endl;
	return true;
}

bool imuse_channel::handleRegion(chunck & src) {
	if(src.get_size() != 8) throw runtime_error("invalid size for REGN chunck");
#ifdef DEBUG_CHANNEL
	int start = from_big_endian(src.get_dword());
	assert(start == _imuse_start);
#else
	src.seek(4);
#endif
	_imuse_length = from_big_endian(src.get_dword());
	//~ assert(_imuse_length == _data_size);
	return true;
}

bool imuse_channel::handleStop(chunck & src) {
	if(src.get_size() != 4) throw runtime_error("invalid size for STOP chunck");
#ifdef DEBUG_CHANNEL
	int _imuse_stop = from_big_endian(src.get_dword());
	assert(_imuse_stop == _imuse_start + _imuse_length);
#endif
	return true;
}

bool imuse_channel::handleMap(chunck & map) {
	while(!map.eof()) {
		chunck * sub = map.sub_block();
		//~ if(sub.get_size() & 1) map.seek(1);
		switch(sub->get_type()) {
			case TYPE_FRMT:
				handleFormat(*sub);
				break;
			case TYPE_TEXT:
				handleText(*sub);
				break;
			case TYPE_REGN:
				handleRegion(*sub);
				break;
			case TYPE_STOP:
				handleStop(*sub);
				break;
			default:
				ostringstream oss;
				oss << "Unknown iMUS subchunck found : " << chunck::chunck_string(sub->get_type()) << ", " << sub->get_size();
				throw runtime_error(oss.str());
		}
		delete sub;
	}
	return true;
}

int imuse_channel::resample_mono(int size) throw(std::exception)
{
	//~ cout << "imuse_channel::resample() _rate == " << _rate << ", _frequency == " << _frequency << endl;
	if(_rate > _frequency) throw runtime_error("Unsupported : imuse rate > output frequency");
	int multiplier = _frequency / _rate;
	//~ cout << "multiplier == " << multiplier << endl;
	if((_rate * multiplier) != _frequency) throw runtime_error("Unsupported : imuse rate is not a multiple of output frequency");
	multiplier *= 2;
	assert(size * multiplier < _frequency*2); // This is because of the _resampled_buffer size
	short * decoded = _decoded_buffer;
	short * resampled = _resampled_buffer;
	if(multiplier == 2) {
		for(int i = 0; i < size; i++) {
			*resampled++ = *decoded;
			*resampled++ = *decoded++;
		}
		return size * 2;
	} else if(multiplier == 4) {
		for(int i = 0; i < size; i++) {
			*resampled++ = *decoded;
			*resampled++ = *decoded;
			*resampled++ = *decoded;
			*resampled++ = *decoded++;
		}
		return size * 4;
	} else {
#ifdef DEBUG_CHANNEL
		cerr << "imuse_channel::resample() multiplier == " << multiplier << endl;
#endif		
		for(int i = 0; i < size; i++) {
			for(int x = 0; x < multiplier; x++) {
				*resampled++ = *decoded;
			}
			decoded++;
		}
		return resampled - _resampled_buffer;
	}
}

int imuse_channel::resample_stereo(int size) throw(std::exception)
{
	assert((size & 1) == 0);
	//~ cout << "imuse_channel::resample() _rate == " << _rate << ", _frequency == " << _frequency << endl;
	if(_rate > _frequency) throw runtime_error("Unsupported : imuse rate > output frequency");
	int multiplier = _frequency / _rate;
	//~ cout << "multiplier == " << multiplier << endl;
	if((_rate * multiplier) != _frequency) throw runtime_error("Unsupported : imuse rate is not a multiple of output frequency");
	if(_channels == 1) multiplier *= 2;
	assert(size * multiplier < _frequency*2); // This is because of the _resampled_buffer size
	short * decoded = _decoded_buffer;
	short * resampled = _resampled_buffer;
	if(multiplier == 1) {
		memcpy(resampled, decoded, size * sizeof(short));
		return size;
	} else if(multiplier == 2) {
		size /= 2;
		for(int i = 0; i < size; i++) {
			short v1 = *decoded++;
			short v2 = *decoded++;
			*resampled++ = v1;
			*resampled++ = v2;
			*resampled++ = v1;
			*resampled++ = v2;
		}
		return size * 4;
	} else {
#ifdef DEBUG_CHANNEL
		cerr << "imuse_channel::resample() multiplier == " << multiplier << endl;
#endif		
		size /= 2;
		for(int i = 0; i < size; i++) {
			short v1 = *decoded++;
			short v2 = *decoded++;
			for(int x = 0; x < multiplier; x++) {
				*resampled++ = v1;
				*resampled++ = v2;
			}
		}
		return resampled - _resampled_buffer;
	}
}

int imuse_channel::decode(int size, int &ret) throw(std::exception)
{
	short * decoded = _decoded_buffer;
	unsigned char * source = _buffer + _data_start;
	//~ cerr << "imuse_channel::decode() bitsize == " << _bitsize << endl;
	switch(_bitsize) {
		case 8:
			while(size--) {
				*decoded++ = (static_cast<short>(*source++) - 128) << 8;
			} break;
		case 12:
			//~ assert((size & 1) == 0);
			{
				int msize = size & 1;
				size = (size+1) >> 1;
				while(size--) {
					int v1 =  *source++;
					int v2 =  *source++;
					int v3 =  *source++;
					int value = (((v2 & 0x0f) << 12) | (v1 << 4)) - 0x8000;
					*decoded++ = static_cast<short>(value);
					value = (((v2 & 0xf0) << 8) | (v3 << 4)) - 0x8000;		
					*decoded++ = static_cast<short>(value);
				} 
				if(msize) {
					decoded--;
#ifdef DEBUG_CHANNEL
					cerr << "imuse_channel::decode() had to skip a sample..." << endl;
#endif
				}
			} break;
		case 16:
			while(size--) {
				int value = ((*source++ << 8) +  *source++) - 0x8000;
				*decoded++ = static_cast<short>(value);
			} break;
		default:
			throw runtime_error("Unsupported bit size in imuse chunck");
	}
	ret = source - (_buffer + _data_start);
	return decoded - _decoded_buffer;
}

bool imuse_channel::processBuffer() throw(exception) {
#ifdef DEBUG_CHANNEL
	cout << "imuse_channel::processBuffer(" << _track << ") : _cur_buffer_write_pos " << _cur_buffer_write_pos << ", _cur_buffer_read_pos " << _cur_buffer_read_pos << endl;
#endif
	while((_cur_buffer_write_pos - _cur_buffer_read_pos) >= 8) {
		int available_size = _cur_buffer_write_pos - _cur_buffer_read_pos;
		chunck::type type = from_big_endian(*reinterpret_cast<unsigned int *>(_buffer + _cur_buffer_read_pos));
		unsigned int size = from_big_endian(*reinterpret_cast<unsigned int *>(_buffer + _cur_buffer_read_pos + 4));
#ifdef DEBUG_CHANNEL
		cout << "TRACK " << _track << ", chunck " << chunck::chunck_string(type) << ", " << available_size << "/" << size+8 << endl;
#endif
		switch(type) {
			case TYPE_MAP_: 
				{
					assert(available_size >= (size + 8));
					cont_chunck map_chunck(reinterpret_cast<char*>(_buffer) + _cur_buffer_read_pos);
					handleMap(map_chunck);
				} break;
			case TYPE_DATA: // Sound data !!!
				if(_data_start == -1) {
					_rdata_start = _data_start = _cur_buffer_read_pos + 8;
					_rdata_size = _data_size = size;
					int reqsize = 1;
					if(_channels == 2) reqsize *= 2;
					if(_bitsize == 16) reqsize *= 2;
					else if(_bitsize == 12) reqsize *= 3;
					if((size % reqsize) != 0)
						throw runtime_error("Invalid sound data size");
				}
				break;
			case TYPE_iMUS:
				size = 0;
				break;
			default:
				{
					ostringstream oss;
					oss << "unknown chunck in iMUS track : " << chunck::chunck_string(type) << endl;
					throw runtime_error(oss.str());
				} break;
		}
		//~ if(_track == 65869) cerr << "imuse_channel::processBuffer(" << _track << ") available_size == " << available_size << ", size == " << size << endl;
		if(available_size >= (size + 8)) {
			_cur_buffer_read_pos += size + 8;
		} else {
			break;
		}
	}
	return true;
}

int imuse_channel::available_sound_data(void) const throw() {
	assert(_data_start <= _cur_buffer_write_pos);
	int size = min(_cur_buffer_write_pos - _data_start, _data_size);
	
	//~ cerr << "imuse_channel::available_sound_data() : size in byte == " << size << endl;
	switch(_bitsize) {
		case 12:
			size = (size / 3) * 2;
			break;
		case 16:
			size /= 2;
	}
	if(_channels == 2)
		size /= 2;
	if(_bitsize == 12) size &= ~1;
	//~ cerr << "imuse_channel::available_sound_data() : size in samples == " << size << endl;
	return size;
}

void imuse_channel::read_sound_data(short * snd, int size) throw(std::exception) {
	if(_data_start == -1 || _data_size == 0) throw(runtime_error("invalid call to saud_channel::read_sound_data()"));
	
#ifdef DEBUG_CHANNEL
	int available =  min(_cur_buffer_write_pos - _data_start, _data_size) * 2;
	switch(_bitsize) {
		case 12:
			available = (available / 3) * 2;
			break;
		case 16:
			available /= 2;
	}
	cout << "saud_channel::read_sound_data(" << _track << ") size == " << size << ", available == " << available << endl;
	assert(available >= size);
#endif
	//~ if(_channels == 2) size *= 2; // We need to decode twice the amount of data
	//~ cerr << "imuse::write() rate == " << _rate << ", bitsize == " << _bitsize << ", channels == " << _channels << endl;
	int resampled_size, really_decoded = 0;
	int decoded_size = decode((_channels == 2) ? size*2 : size, really_decoded);
	_data_start += really_decoded;
	_data_size -= really_decoded;
	if(_channels == 2)
		resampled_size = resample_stereo(decoded_size); 
	else
		resampled_size = resample_mono(decoded_size);
#ifdef DEBUG_CHANNEL
	cerr << "imuse::write() size * 2 == " << size * 2 << ", resampled_size == " << resampled_size << endl;
#endif
	//~ if(_bitsize != 12 || _channels != 1)
	for(int i = 0; i < size * 2; i++)
		snd[i] += _resampled_buffer[i];
}



