/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __CHANNEL_H_
#define __CHANNEL_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef DEBUG
# ifndef NO_DEBUG_CHANNEL
#  define DEBUG_CHANNEL
# endif
#else
# ifdef DEBUG_CHANNEL
#  error DEBUG_CHANNEL defined without DEBUG
# endif
#endif

#include <exception>

class chunck;
class cont_chunck;
	
/*! 	@brief interface for a sound channel (a track)

	This is the interface for sound channels. 
*/
class channel {
public:
	virtual ~channel() throw() {};
	/*! 	@brief is the track terminated ?
	
		This method is called by the mixer from time to time to check if the track is terminated.
		
		@remark If a track has no more available data, it does not necesseraly means that it is terminated.
		
		@return true if the sound channel is terminated, false otherwise
	*/
	virtual bool is_terminated() const throw() = 0;
	/*! 	@brief appends coded sound data to the track.
	
		This method is called by the player from time to time when new data is available for the track.
		
		@param b the chunck to read data from.
		@param size the size of the data to read.
		
		@return true if everything went fine, false otherwise
	*/
	virtual bool append_data(chunck & b, int size) throw(std::exception) = 0;
	/*! 	@brief get the number of samples currently available in the channel.
	
		This method is called by the mixer from time to time to get the current number of samples available.
		
		@remark If a track has no more available data, it does not necesseraly means that it is terminated.
		
		@return the number of samples available
	*/
	virtual int available_sound_data() const throw() = 0;
	/*! 	@brief read some sound data from the channel.
	
		This method is called by the mixer from time to time to get the current number of samples available.

		@param sound_buffer the buffer to write data to.
		@param size the size of the data to write (in samples).
		
		@remark The sound_buffer may already has been set by another track, so the track should add to it.
		
		@return the number of samples available
	*/
	virtual void read_sound_data(short * sound_buffer, int size) throw(std::exception) = 0;
	/*! 	@brief debugging do not use.
	
		This method is called by player to perform integrity check on the unknown parameters in the
		chuncks. There is no real other point than to check a few assertions in using this method.
	*/
	virtual bool set_parameters(int, int, int, int) throw(std::exception) = 0;
	/*! 	@brief debugging do not use.
	
		This method is called by player to perform integrity check on the unknown parameters in the
		chuncks. There is no real other point than to check a few assertions in using this method.
	*/
	virtual bool check_parameters(int, int, int, int, int) throw(std::exception) = 0;
};

/*! 	@brief class for a SAUD sound ::channel (a Full Throttle track)

	This class implements a channel specifically for Full Throttle.
*/
class saud_channel : public channel {
private:
	int _track;				//!< The track identifier
	int _nbframes;			//!< number of frames of the track (unused)
	unsigned char * _buffer;	//!< SAUD buffer
	int _buffer_size;			//!< the size of the sound buffer
	int _frequency;			//!< the frequency target of the track (always 22050)
	int _cur_buffer_write_pos;	//!< the current write position in the SAUD buffer
	int _cur_buffer_read_pos;	//!< the current read position in the SAUD buffer
	int _data_start;			//!< the start of the sound data in the SAUD buffer 
	int _data_size;			//!< remaining size of sound data in the SAUD buffer
	bool _mark_reached;		//!< set to \c true when the SMRK tag is reached
	int _flags;				//!< current flags of the track (unused)
	int _volume;			//!< the current track volume
	int _balance;			//!< the current track balance
	int _index;				//!< the current PSAD index (for coherency checking)
	short _voltable[2][256];	//!< the precalculated volume table (stereo 16 bits)
	int _rdata_start;			//!< a copy of \c _data_start that is not modified by \c read_sound_data(). Used to save the sound data to a file.
	int _rdata_size;			//!< a copy of \c _data_size that is not modified by \c read_sound_data(). Used to save the sound data to a file.
protected:
	/*!	@brief STRK chunck handling

		This method contain the handling of the STRK sub-chunck.

		@param c the STRK chunck to handle.
	*/
	void handleStrk(chunck & c) throw(std::exception);
	/*!	@brief SMRK chunck handling

		This method contain the handling of the SMRK sub-chunck.

		@param c the SMRK chunck to handle.
	*/
	void handleSmrk(chunck & c) throw(std::exception);
	/*!	@brief SHDR chunck handling

		This method contain the handling of the SHDR sub-chunck.

		@param c the SHDR chunck to handle.
	*/
	void handleShdr(chunck & c) throw(std::exception);
	/*!	@brief SAUD buffer processing.

		this method is called by \c append_data to check if something can be done from the available data.
		
		@return true if everything went fine, false otherwise
	*/
	bool processBuffer() throw(std::exception);
	/*!	@brief volume table calculation.

		This method is called when the \c _volume or the \c _balance change.
		This method fills the \c _voltable tables.
	*/
	void recalcVolumeTable() throw(std::exception);
public:
	saud_channel(int track, int freq) throw();
	virtual ~saud_channel() throw();
	bool is_terminated() const throw();
	bool set_parameters(int duration, int flags, int vol1, int vol2) throw(std::exception);
	bool check_parameters(int index, int duration, int flags, int vol1, int vol2) throw(std::exception);
	bool append_data(chunck & b, int size) throw(std::exception);
	int available_sound_data() const throw();
	void read_sound_data(short * sound_buffer, int size) throw(std::exception);
};

/*! 	@brief class for a IACT sound ::channel (a The Dig track)

	This class implements a channel specifically for The Dig.

	\bug for unknown reason, some sound have a too long duration, or repeat themselves.
*/
class imuse_channel : public channel {
private:
	int _track;				//!< the track number
	unsigned char * _buffer;	//!< the buffer that contain the iMUS chunck
	short * _decoded_buffer;	//!< a buffer for the decoded sound data
	short * _resampled_buffer;	//!< a buffer for the resampled sound data
	int _buffer_size;			//!< the size of the iMUS buffer
	int _frequency;			//!< the target frequency of the ::mixer
	int _cur_buffer_write_pos;	//!< the current write position in the iMUS buffer
	int _cur_buffer_read_pos;	//!< the current read position in the iMUS buffer
	int _data_start;			//!< the start of the sound data in the iMUS buffer 
	int _data_size;			//!< remaining size of sound data in the iMUS buffer
	int _size;
	int _nbframes;
	int _unk1;
	int _unk2;
	int _index;
	int _bitsize;			//!< the bitsize of the original data
	int _rate;				//!< the sampling rate of the original data
	int _channels;			//!< the number of channels of the original data
	int _imuse_start;
	int _imuse_length;
	int _imuse_stop;
	int _rdata_start;			//!< a copy of \c _data_start that is not modified by \c read_sound_data(). Used to save the sound data to a file.
	int _rdata_size;			//!< a copy of \c _data_size that is not modified by \c read_sound_data(). Used to save the sound data to a file.
protected:
	int resample_mono(int size) throw(std::exception);
	int resample_stereo(int size) throw(std::exception);
	int decode(int size, int &) throw(std::exception);
	bool processBuffer() throw(std::exception);
	bool handleMap(chunck &);
	bool handleFormat(chunck &);
	bool handleText(chunck &);
	bool handleRegion(chunck &);
	bool handleStop(chunck &);
public:
	imuse_channel(int track, int freq) throw();
	virtual ~imuse_channel() throw();
	bool is_terminated() const throw();
	bool set_parameters(int nbframes, int size, int unk1, int unk2) throw(std::exception);
	bool check_parameters(int index, int nbframes, int size, int unk1, int unk2) throw(std::exception);
	bool append_data(chunck & b, int size) throw(std::exception);
	int available_sound_data() const throw();
	void read_sound_data(short * sound_buffer, int size) throw(std::exception);
};

#endif
