/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#include <sstream>
#include <fstream>
#include <stdexcept>
#include "endian.h"
#include "chunck.h"

//~ #include <iostream>

using namespace std;

/*!	@brief very small and fast wrapper for a ifstream.

	@todo implement reference counting, so that ::file_chunck does not leak memory !
*/
class file_ptr {
	ifstream _ifs;
public:
	file_ptr(const string & fname) {
		_ifs.open(fname.c_str(), ios::binary);
	}
	~file_ptr() {
		_ifs.close();
	}
	int tell() {
		return _ifs.tellg();
	}
	bool seek(int pos) {
		_ifs.seekg(pos);
		return true;
	}
	bool read(void * ptr, int size) {
		_ifs.read(static_cast<char*>(ptr), size);
		return true;
	}
};

string chunck::chunck_string(chunck::type t) {
	ostringstream oss;
	oss << "TYPE:" 
		<< static_cast<char>((t >> 24) & 0xFF) 
		<< static_cast<char>((t >> 16) & 0xFF) 
		<< static_cast<char>((t >> 8) & 0xFF) 
		<< static_cast<char>(t & 0xFF);
	return oss.str();
}


file_chunck::file_chunck() throw() : _data(0), _type(0), _size(0), _cur_pos(0) {
}

//~ chunck::chunck(const chunck & b) throw() : _data(b._data), _type(b._type), _size(b._size), _offset(b._offset), _cur_pos(0) {
//~ }

file_chunck::file_chunck(const string & fname) throw(exception) {
	_data = new file_ptr(fname);
	_data->read(&_type, 4);
	_type = from_big_endian(_type);
	_data->read(&_size, 4);
	_size = from_big_endian(_size);
	_offset = _data->tell();
	_cur_pos = 0;
}

//~ chunck & chunck::operator=(const chunck& b) {
	//~ _data = b._data;
	//~ _type = b._type;
	//~ _size = b._size;
	//~ _offset = b._offset;
	//~ _cur_pos = 0;
	//~ return *this;
//~ }

chunck::type file_chunck::get_type() const throw() { 
	return _type; 
}

unsigned int file_chunck::get_size() const throw() { 
	return _size; 
}

chunck * file_chunck::sub_block() throw(exception) {
	file_chunck * ptr = new file_chunck;
	ptr->_data = _data;
	_data->seek(_offset+_cur_pos);
	unsigned int temp;
	_data->read(&temp, 4);
	ptr->_type = from_big_endian(temp);
	_data->read(&temp, 4);
	ptr->_size = from_big_endian(temp);
	ptr->_offset = _offset + _cur_pos + 8;
	ptr->_cur_pos = 0;
	//~ cout << chunck_string(ptr._type) << ", " << ptr._size << endl;
	seek(8 + ptr->get_size());
	return ptr;
}

bool file_chunck::eof() const throw() { 
	return _cur_pos >= _size; 
}

unsigned int file_chunck::tell() const throw() { 
	return _cur_pos; 
}

bool file_chunck::seek(int delta, ios::seekdir dir) throw (exception) {
	switch(dir) {
		case ios::cur:
			_cur_pos += delta;
			break;
		case ios::beg:
			if(delta < 0) throw runtime_error("invalid seek request");
			_cur_pos = static_cast<unsigned int>(delta);
			break;
		case ios::end:
			if(delta > 0 || (_size + delta) < 0) throw runtime_error("invalid seek request");
			_cur_pos = static_cast<unsigned int>(_size + delta);
			break;
	}
	if(_cur_pos > _size) {
		ostringstream oss;
		oss << "invalid seek request : " << _cur_pos << " > " << _size << "(delta == " << delta << ")";
		throw runtime_error(oss.str());
	}
	return true;
}

bool file_chunck::read(void * buffer, unsigned int size) throw(exception) {
	if(size <= 0 || (_cur_pos + size) > _size) throw runtime_error("invalid buffer read request");
	_data->seek(_offset + _cur_pos);
	_data->read(buffer, size);
	_cur_pos += size;
	return true;
}

char file_chunck::get_char() throw(exception) {
	if(_cur_pos >= _size) throw runtime_error("invalid char read request");
	_data->seek(_offset + _cur_pos);
	char buffer;
	_data->read(&buffer, sizeof(buffer));
	_cur_pos+= sizeof(buffer);
	return buffer;
}

unsigned char file_chunck::get_byte() throw(exception) {
	if(_cur_pos >= _size) throw runtime_error("invalid byte read request");
	_data->seek(_offset + _cur_pos);
	unsigned char buffer;
	_data->read(&buffer, sizeof(buffer));
	_cur_pos+= sizeof(buffer);
	return buffer;
}

short file_chunck::get_short() throw(exception) {
	if(_cur_pos >= _size-1) throw runtime_error("invalid short read request");
	_data->seek(_offset + _cur_pos);
	short buffer;
	_data->read(&buffer, sizeof(buffer));
	_cur_pos+= sizeof(buffer);
	return buffer;
}

unsigned short file_chunck::get_word() throw(exception) {
	if(_cur_pos >= _size-1) throw runtime_error("invalid word read request");
	_data->seek(_offset + _cur_pos);
	unsigned short buffer;
	_data->read(&buffer, sizeof(buffer));
	_cur_pos+= sizeof(buffer);
	return buffer;
}

unsigned int file_chunck::get_dword() throw(exception) {
	if(_cur_pos >= _size-3) throw runtime_error("invalid dword read request");
	//~ unsigned int * ptr = reinterpret_cast<unsigned int *>(_data + _cur_pos);
	//~ _cur_pos += 4;
	//~ return *ptr;
	_data->seek(_offset + _cur_pos);
	unsigned int buffer;
	_data->read(&buffer, sizeof(buffer));
	_cur_pos+= sizeof(buffer);
	return buffer;
}

//~ cont_chunck::cont_chunck() throw() : _data(0), _type(0), _size(0), _cur_pos(0) {
//~ }

//~ cont_chunck::cont_chunck(const cont_chunck & b) throw() : _data(b._data), _type(b._type), _size(b._size), _cur_pos(0) {
//~ }

cont_chunck::cont_chunck(char * data) throw(exception) {
	if(data == 0) throw(logic_error("chunck() called with NULL pointer"));
	_type = from_big_endian(reinterpret_cast<chunck::type*>(data)[0]);
	_size = from_big_endian(reinterpret_cast<unsigned int*>(data)[1]);
	_data = data + sizeof(chunck::type) + sizeof(unsigned int);
	_cur_pos = 0;
}

//~ cont_chunck & cont_chunck::operator=(const cont_chunck& b) {
	//~ _data = b._data;
	//~ _type = b._type;
	//~ _size = b._size;
	//~ _cur_pos = 0;
	//~ return *this;
//~ }

chunck::type cont_chunck::get_type() const throw() { 
	return _type; 
}

unsigned int cont_chunck::get_size() const throw() { 
	return _size; 
}

chunck * cont_chunck::sub_block() throw(exception) {
	cont_chunck * ptr = new cont_chunck(_data + _cur_pos);
	seek(sizeof(chunck::type) + sizeof(unsigned int) + ptr->get_size());
	return ptr;
}

bool cont_chunck::eof() const throw() { 
	return _cur_pos >= _size; 
}

unsigned int cont_chunck::tell() const throw() { 
	return _cur_pos; 
}

bool cont_chunck::seek(int delta, ios::seekdir dir) throw (exception) {
	switch(dir) {
		case ios::cur:
			_cur_pos += delta;
			break;
		case ios::beg:
			if(delta < 0) throw runtime_error("invalid seek request");
			_cur_pos = static_cast<unsigned int>(delta);
			break;
		case ios::end:
			if(delta > 0 || (_size + delta) < 0) throw runtime_error("invalid seek request");
			_cur_pos = static_cast<unsigned int>(_size + delta);
			break;
	}
	if(_cur_pos > _size) {
		ostringstream oss;
		oss << "invalid seek request : " << _cur_pos << " > " << _size << "(delta == " << delta << ")";
		throw runtime_error(oss.str());
	}
	return true;
}

bool cont_chunck::read(void * buffer, unsigned int size) throw(exception) {
	if(size <= 0 || (_cur_pos + size) > _size) throw runtime_error("invalid buffer read request");
	memcpy(buffer, _data + _cur_pos, size);
	_cur_pos += size;
	return true;
}

char cont_chunck::get_char() throw(exception) {
	if(_cur_pos >= _size) throw runtime_error("invalid char read request");
	return _data[_cur_pos++];
}

unsigned char cont_chunck::get_byte() throw(exception) {
	if(_cur_pos >= _size) throw runtime_error("invalid byte read request");
	unsigned char * ptr = reinterpret_cast<unsigned char *>(_data + _cur_pos);
	_cur_pos += 1;
	return *ptr;
}

short cont_chunck::get_short() throw(exception) {
	if(_cur_pos >= _size-1) throw runtime_error("invalid short read request");
	short * ptr = reinterpret_cast<short *>(_data + _cur_pos);
	_cur_pos += 2;
	return *ptr;
}

unsigned short cont_chunck::get_word() throw(exception) {
	if(_cur_pos >= _size-1) throw runtime_error("invalid word read request");
	unsigned short * ptr = reinterpret_cast<unsigned short *>(_data + _cur_pos);
	_cur_pos += 2;
	return *ptr;
}

unsigned int cont_chunck::get_dword() throw(exception) {
	if(_cur_pos >= _size-3) throw runtime_error("invalid dword read request");
	unsigned int * ptr = reinterpret_cast<unsigned int *>(_data + _cur_pos);
	_cur_pos += 4;
	return *ptr;
}
