/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __CHUNCK_H_
#define __CHUNCK_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string>
#include <exception>
#include <ostream>
/*! 	@brief Interface for chunck handling

	This class is an interface for reading from a chunck.

	\todo handle big endian system.
*/
class chunck {
public:
	virtual ~chunck() throw() {};
	typedef unsigned int type;			//!< type of a chunck (i.e. The first 4byte field of the chunck structure).
	/*!	@brief convert a type to a string
		
		Utility function that convert a type to a string.
		
		@param t the type to convert to a string
		
		@return the converted string
	*/
	static std::string chunck_string(type t);

	virtual type get_type() const throw() = 0;	//!< return the type of the chunck
	virtual unsigned int get_size() const throw() = 0;	//!< return the size of the chunck
	virtual chunck * sub_block() throw(std::exception) = 0; //!< extract a subchunck from the current read position
	virtual bool eof() const throw() = 0;	//!< is the chunck completely read ?
	virtual unsigned int tell() const throw() = 0;	//!< get the chunck current read position
	virtual bool seek(int delta, std::ios::seekdir dir = std::ios::cur) throw (std::exception) = 0;	//!< move the current read position inside the chunck
	virtual bool read(void * buffer, unsigned int size) throw(std::exception) = 0;		//!< read some data for the current read position
	virtual char get_char() throw(std::exception) = 0;							//!< extract the character at the current read position
	virtual unsigned char get_byte() throw(std::exception) = 0;					//!< extract the byte at the current read position
	virtual short get_short() throw(std::exception) = 0;						//!< extract the short at the current read position
	virtual unsigned short get_word() throw(std::exception) = 0;					//!< extract the word at the current read position
	virtual unsigned int get_dword() throw(std::exception) = 0;					//!< extract the dword at the current read position
};

class file_ptr;

/*! 	@brief file based ::chunck

	This class is an implementation of ::chunck that handles file.

*/
class file_chunck : public chunck {
private:
	file_ptr * _data;
	type _type;
	unsigned int _size;
	unsigned int _offset;
	unsigned int _cur_pos;
protected:
	file_chunck() throw();
public:
	file_chunck(const std::string & fname) throw(std::exception);
	type get_type() const throw();
	unsigned int get_size() const throw();
	chunck * sub_block() throw(std::exception);
	bool eof() const throw();
	unsigned int tell() const throw();
	bool seek(int delta, std::ios::seekdir dir = std::ios::cur) throw (std::exception);
	bool read(void * buffer, unsigned int size) throw(std::exception);
	char get_char() throw(std::exception);
	unsigned char get_byte() throw(std::exception);
	short get_short() throw(std::exception);
	unsigned short get_word() throw(std::exception);
	unsigned int get_dword() throw(std::exception);
};

/*! 	@brief memory based ::chunck

	This class is an implementation of ::chunck that handles a memory buffer.
*/
class cont_chunck : public chunck {
private:
	char * _data;
	chunck::type _type;
	unsigned int _size;
	unsigned int _cur_pos;
public:
	cont_chunck(char * data) throw(std::exception);
	chunck::type get_type() const throw();
	unsigned int get_size() const throw();
	chunck * sub_block() throw(std::exception);
	bool eof() const throw();
	unsigned int tell() const throw();
	bool seek(int delta, std::ios::seekdir dir = std::ios::cur) throw (std::exception);
	bool read(void * buffer, unsigned int size) throw(std::exception);
	char get_char() throw(std::exception);
	unsigned char get_byte() throw(std::exception);
	short get_short() throw(std::exception);
	unsigned short get_word() throw(std::exception);
	unsigned int get_dword() throw(std::exception);
};

#endif
