/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#include <cstring>
#ifdef DEBUG_FONT_RENDERER
#include <iostream>
#endif
#include "frenderer.h"
#include "rect.h"

using namespace std;

font_renderer::font_renderer(bool use_original_colors) throw() : _nb_frames(0), _color(-1), _original(use_original_colors)
{
}

font_renderer::~font_renderer() throw() {
	for(vector<char *>::iterator i = _chars.begin(); i != _chars.end(); i++) {
		delete [](*i);
	}
}

void font_renderer::save(int frame) {
	_nb_frames++;
	_sizes.push_back(make_pair(get_width(), get_height()));
	int size = get_width()*get_height();
	char * copy = new char[size];
	memcpy(copy, data(), size);
	_chars.push_back(copy);
}

int font_renderer::char_width(int v) const throw(exception) {
	if(v < 0 || v > _nb_frames) throw(runtime_error("invalid character in font_renderer::char_size"));
	return _sizes[v].first;
}

int font_renderer::char_height(int v) const throw(exception) {
	if(v < 0 || v > _nb_frames) throw(runtime_error("invalid character in font_renderer::char_size"));
	return _sizes[v].second;
}

int font_renderer::string_width(const string & str) const throw(exception) {
	int ret = 0;
	string::const_iterator i = str.begin();
	while(i != str.end()) {
		ret += char_width(*i++);
	}
	return ret;
}

int font_renderer::string_height(const string & str) const throw(exception) {
	int ret = 0;
	string::const_iterator i = str.begin();
	while(i != str.end()) {
		ret = max(ret, char_height(*i++));
	}
	return ret;
}

int font_renderer::draw_char(char * buffer, const point & size, int x, int y, int chr) const throw(exception) {
	int w = _sizes[chr].first;
	int h = _sizes[chr].second;
	//~ cout << "draw_char(" << chr << ", " << pos << " == " << chsize << endl;
	char * src = _chars[chr];
	char * dst = buffer + size.get_x()*y+x;
	if(_original) {
		for(int j = 0; j < h; j++) {
			for(int i = 0; i < w; i++) {
				int value = *src++;
				if(value) dst[i] = value;
			}
			dst += size.get_x();
		}
	} else {
		int color = (_color != -1) ? _color : 1;
		for(int j = 0; j < h; j++) {
			for(int i = 0; i < w; i++) {
				int value = *src++;
				if(value == 1) {
					//~ cout << pos.get_x()+i << ", " << pos.get_y()+j << endl;
					dst[i] = color;
				} else if(value) {
					dst[i] = 0;
				}
			}
			dst += size.get_x();
		}
	}
	return w;
}

namespace {
	vector<string> split(const string & str, char sep) {
		vector<string> ret;
		string::size_type i = 0, j = str.find(sep);
		while(i != string::npos) {
			ret.push_back(string(str, i, j - i));
			if(j != string::npos) {
				i = j+1;
				j = str.find(sep, i);
			} else 
				i = j;
		}
		//~ cout << "split == " << ret.size() << endl;
		//~ for(int i = 0; i < ret.size(); i++) {
			//~ cout << "split[" << i << "] == " << ret[i] << endl;
		//~ }
		return ret;
	}
}

void font_renderer::draw_substring(const std::string & str, char * buffer, const point & size, int x, int y) const throw(exception)
{
#ifdef DEBUG_FONT_RENDERER
	cout << "font_renderer::draw_substring(" << str << "), x == " << x << ", y == " << y << endl;
#endif
	for(string::const_iterator i = str.begin(); i != str.end(); i++)
		x += draw_char(buffer, size, x, y, *i);
}

bool font_renderer::draw_string_absolute(const std::string & str, char * buffer, const point & size, int x, int y) const throw(exception) {
	draw_substring(str, buffer, size, x, y);
	return true;
}

bool font_renderer::draw_string_centered(const std::string & str, char * buffer, const point & size, int y, int xmin, int width, int offset) const throw(exception) {
	//~ cout << "font_renderer::draw_string(" << str << ") == " << r << endl;
	vector<string> words = split(str, ' ');
	vector<int> sizes(words.size());
	for(int i = 0; i < words.size(); i++)
		sizes[i] = string_width(words[i]);
	int i = 0, max_width = 0, height = 0;
	vector<string> substrings;
	vector<int> substr_widths;
	vector<int> substr_heights;
	int space_width = char_width(' ');
	while(i < sizes.size()) {
		int substr_width = sizes[i];
		string substr = words[i];
		int j = i + 1;
		while(j < sizes.size() && (substr_width + space_width + sizes[j]) < width) {
			substr_width += sizes[j++] + space_width;
		}
		for(int k = i +1; k < j; k++) {
			substr += " ";
			substr += words[k];
		}
		substrings.push_back(substr);
		substr_widths.push_back(substr_width);
		if(substr_width > max_width)
			max_width = substr_width;
		i = j;
		height += string_height(substr);
	}
	max_width = (max_width + 1) >> 1;
	// we have a box from 0 -> max_width
	// we want a box from (xmin + offset) - max_width / 2, (xmin + offset) + max_width / 2
	int x = xmin + width / 2;
	x += offset - size.get_x() / 2;
//	*/ + max_width;
	if(x < max_width) x = max_width;
	if(x + max_width > size.get_x()) {
		x = size.get_x() - max_width;
	}
	if(y + height > size.get_y()) {
		y = size.get_y() - height;
	}
	for(i = 0; i < substrings.size(); i++) {
		int substr_width = substr_widths[i];
		draw_substring(substrings[i], buffer, size, x - substr_width / 2, y);
		y += string_height(substrings[i]);
	}
	return true;
}
