/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __FRENDERER_H_
#define __FRENDERER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef DEBUG
# ifndef NO_DEBUG_FONT_RENDERER
#  define DEBUG_FONT_RENDERER
# endif
#else
# ifdef DEBUG_FONT_RENDERER
#  error DEBUG_FONT_RENDERER defined without DEBUG
# endif
#endif

#include <vector>
#include <string>
#include <stdexcept>
#include "brenderer.h"
#include "rect.h"
#include "blitter.h"

/*!	@brief ::renderer implementation specifically designed for font files.

	This class is a valid ::renderer implementation. The frames are kept in memory, as bitmap representing characters, so that 
	they can be rendered again in another frame as strings. 
	
	This class also contains some functions useful for printing strings. This is used to show subtitles and more generally texts
	in animations.
	
	@todo update the mehod to use the ::blitter class, instead of direct pointers.
*/
class font_renderer : public base_renderer {
private:
	int _nb_frames;	//!< The number of frames in the font
	int _color;		//!< A color parameter used for font printing.
	bool _original;	//!< flag for color selection
	std::vector<std::pair<int, int> > _sizes;	//!< array that contains the size of the different frames (i.e. characters) of the font.
	std::vector<char *> _chars;			//!< array of the font data.
public:
	/*!	@brief font_renderer constructor
		
		@param use_original_colors	flag to indicate if the font use it's own color, or if the base color are set at runtime.
	*/
	font_renderer(bool use_original_colors = false) throw();
	virtual ~font_renderer() throw();
	virtual bool wait(int ms) throw() { return true; };
protected:
	virtual void save(int frame = -1);
	/*!	@brief get the width of a character.
		
		@param c	the character we want the width from.

		@return the width of the character
	*/
	int char_width(int c) const throw(std::exception);
	/*!	@brief get the width of a string.
		
		@param str	the string we want the width from.

		@return the complete width of the string
	*/
	int string_width(const std::string & str) const throw(std::exception);
	/*!	@brief get the height of a character.
		
		@param c	the character we want the height from.

		@return the height of the character
	*/
	int char_height(int c) const throw(std::exception);
	/*!	@brief get the height of a string.
		
		@param str	the string we want the height from.

		@return the complete height of the string
	*/
	int string_height(const std::string & str) const throw(std::exception);
	/*!	@brief draw a character in the given frame buffer.
		
		@param buffer	the frame buffer to draw into.
		@param size		the size of the frame buffer.
		@param x		the horizontal position of the topleft corner of the character.
		@param y		the vertical position of the topleft corner of the character.
		@param c		the character to draw.
		
		@bug	This method does not clip. This is not really a bug, as it should always be correctly called, but some asserts would be welcome.

		@return the width of the character
	*/
	int draw_char(char * buffer, const point & size, int x, int y, int c) const throw(std::exception);
	/*!	@brief draw a string in the given frame buffer.
		
		@param str		the string to draw.
		@param buffer	the frame buffer to draw into.
		@param size		the size of the frame buffer.
		@param x		the horizontal position of the topleft corner of the string.
		@param y		the vertical position of the topleft corner of the string.

		@bug	This method does not clip. This is not really a bug, as it should always be correctly called, but some asserts would be welcome.
	*/
	void draw_substring(const std::string & str, char * buffer, const point & size, int x, int y) const throw(std::exception);
public:
	/*!	@brief change the programmable color of the font.
		
		@param c		the new color to use.

		@return \c true if everything went fine, \c false otherwise
	*/
	bool set_color(int c) throw() { _color = c; return true; }
	/*!	@brief draw a centered and possibly using multiple lines string.
	
		This method performs calculation of the string size before choosing where to draw it.
		As I still not have figured out exactly what is the meaning of the fields in the TRES chunck, 
		the real meaning of the parameters can be quite difficult to understand.
	
		@remark	The current implementation is incorrect in the sense that it does not conform to the original game.
		@todo	rewrite and rethink this to better match the original implementation.
		
		@param str		the string to draw.
		@param buffer	the frame buffer to draw into.
		@param size		the size of the frame buffer.
		@param y		the vertical position of the topleft corner of the string. This position may be changed if it is too low to be correctly drawn.
		@param xmin		the minimum horizontal position of the topleft corner of the string.
		@param width		the maximum width of the string. If the string is too long, it will wrap.
		@param offset	offset to give to the horizontal position.

		@return \c true if everything went fine, \c false otherwise
	*/
	bool draw_string_centered(const std::string & str, char * buffer, const point & size, int y, int xmin, int width, int offset) const throw(std::exception);
	/*!	@brief draw a string at an absolute position.
	
		@param str		the string to draw.
		@param buffer	the frame buffer to draw into.
		@param size		the size of the frame buffer.
		@param x		the horizontal position of the topleft corner of the string.
		@param y		the vertical position of the topleft corner of the string. This position may be changed if it is too low to be correctly drawn.

		@return \c true if everything went fine, \c false otherwise
	*/
	bool draw_string_absolute(const std::string & str, char * buffer, const point & size, int x, int y) const throw(std::exception);
};

#endif
