/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#include <iostream>
#include <string>
#include <vector>

#include "player.h"
#include "win32.h"

using namespace std;

int main(int argc, char * argv[]) {
	renderer * render = 0;
	bool wait = true;
	bool sound = true;
	string anim;
	vector<string> show_opt, hide_opt;
	for(int i = 1; i < argc; i++)
	{
		string arg = argv[i];
		if(arg == "-wnd") {
			if(render) {
				cerr << "only one of -wnd | -bmp supported" << endl;
				return 1;
			}
			render = new window_renderer;
		} else if(arg == "-nosound") {
			sound = false;
		} else if(arg == "-show") {
			if(i >= argc-1) { cerr << "-show need an option" << endl; return 1; }
			show_opt.push_back(argv[++i]);
		} else if(arg == "-hide") {
			if(i >= argc-1) { cerr << "-show need an option" << endl; return 1; }
			hide_opt.push_back(argv[++i]);
		} else if(arg == "-nowait") {
			sound = false;
			wait = false;
		} else if(arg == "-bmp") {
			string param = i < argc-1 ? argv[++i] : "frame";
			if(render) {
				cerr << "only one of -wnd | -bmp supported" << endl;
				return 1;
			}
			render = new bitmap_file_renderer(param);
		} else {
			if(anim.size()) {
				cerr << "only one SMUSH file at a time, please" << endl;
				return 1;
			}
			anim = arg;
		}
	}
	if(!anim.size()) {
		cout << "Usage : " << argv[0] << " [<options> ...] <file>" << endl;
		return 1;
	}
	if(!render)
		render = new null_renderer;
	player p(render, wait, sound);
	for(vector<string>::iterator i = show_opt.begin(); i != show_opt.end(); i++)
		p.show(*i);
	for(vector<string>::iterator i = hide_opt.begin(); i != hide_opt.end(); i++)
		p.hide(*i);
	set_terminate(__gnu_cxx::__verbose_terminate_handler);
	try {
		p.play(anim);
	} catch(const exception & e) {
		cerr << "Exception caught : " << e.what() << endl;
	}
	delete render;
	return 0;
}

/*! 	\mainpage SMUSH Player

This is the source code release of a SMUSH Player.

\sa \ref quickstart
\sa \ref licinfo

	\author Xavier Trochu
	\version 1.0
	\date    08-08-2002
*/
/*! \page quickstart A quick introduction to the SMUSH Player

\section quickintro What is this software for ?

This sofware can be used to play the animation files for the following LucasArts Entertainment Company games :
- Full Throttle
- The Dig

This software has been written at the same time as was written the documentation for the SMUSH file format.

I wrote this software because my most beloved game from LEC is Full Throttle, and because <a href="http://scummvm.sourceforge.net/">scummvm</a>
does not support this game yet.

\section quickstruct Structure of the SMUSH Player.

I tried to used an object oriented design when I wrote this software.

The player itself is implemented in a specific class. The output (i.e. frames and sound) is then sent to a renderer, which is dependant on what we want
to do. 

\sa ::player
\sa ::renderer
\sa ::sound_renderer

\section quicksync Synchronisation of Sound and Video.

I haven't found yet if there was a way to perfectly synchronise video and sound in the original format, so right now I use the following idea, which works great with
Full Throttle (it's not as good with The Dig).

When a frame is terminated, The sound mixer enters a loop where it tries to output as much sound data as possible. For this, it iterates within all the active tracks and query them
for their available size (look in mixer::handleFrame()). The loop is exited when either a track as no more data available, or there is no more active track.
Then, the player will wait for all the sound data from the previous frame to have been played before starting to decode the next frame.

\sa <a href="../../Documentation_for_SMUSH_files.html">Documentation for the SMUSH File Format</a>

*/
/*!	\page licinfo License Information
\section copy Copyright information
Copyright (&copy;) 2002 Xavier Trochu
\section lic Licensing information
This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:

-# The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
-# Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
-# This notice may not be removed or altered from any source distribution.
*/

