/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#include <sstream>
#include <stdexcept>
#include <map>
#include "mixer.h"
#include "channel.h"
#include "renderer.h"

#ifdef DEBUG_MIXER
#include <iostream>
#endif

using namespace std;

mixer::mixer() throw() : _frequency(0), _sound(0), _out(0) {
}

bool mixer::init(sound_renderer * sr, int freq, int fps) throw() {
	_frequency = freq;
	_sound = new short[_frequency*2]; // This is too much, but who cares.
	if(_sound == 0) return false;
	_out = sr;
	_out->init(_frequency, true);
	return true;
}

bool mixer::stop() throw(exception) {
	while(handle_frame());
	_out->stop();
	return true;
}

mixer::~mixer() throw() {
	while(_channels.size()) {
		map<int, channel *>::iterator i = _channels.begin();
		delete i->second;
		_channels.erase(i);
	}
	if(_sound) delete []_sound;
}

channel * mixer::find_channel(int track) throw() {
	assert(_frequency != 0);
	map<int, channel *>::iterator i = _channels.find(track);
	return i == _channels.end() ? 0 : i->second;
}

bool mixer::add_channel(int track, channel * c) throw(exception) {
	assert(_frequency != 0);
#ifdef DEBUG_MIXER
	map<int, channel *>::iterator i = _channels.find(track);
	if(i != _channels.end()) throw runtime_error("channel already exist !");
#endif
	_channels.insert(make_pair(track, c));
	return true;
}

bool mixer::handle_frame() throw(std::exception) {
	int dumped_size = 2205;
	while(dumped_size > 0) {
		//~ cout << "mixer::handle_frame() dumped_size == " << dumped_size << endl;
		if(_channels.size() == 0) {
#ifdef DEBUG_MIXER
			cout << "mixer::handle_frame() : no more track available and still " << dumped_size << " bytes to fill" << endl;
#endif
			memset(_sound, 0, 2205 * 4);	
			_out->write(reinterpret_cast<unsigned short *>(_sound), 2205 * 4);
			return false;
		}
		for(map<int, channel *>::iterator i = _channels.begin(); i != _channels.end(); i++) {
			if(i->second->is_terminated()) {
				delete i->second;
				_channels.erase(i);
			}
		}
		int minsize = dumped_size;
		for(map<int, channel *>::iterator i = _channels.begin(); i != _channels.end(); i++) {
			int s = i->second->available_sound_data();
			if(s < minsize) minsize = s;
		}
		//~ minsize &= ~1;
		if(minsize == 0) {
#ifdef DEBUG_MIXER
			cout << "mixer::handle_frame() : no more data in track available and still " << dumped_size << " samples to fill" << endl;
#endif
			return true;
		//~ } else {
			//~ cout << "mixer::handle_frame() minsize == " << minsize << endl;
		}
		memset(_sound, 0, minsize * 4);	
		for(map<int, channel *>::iterator i = _channels.begin(); i != _channels.end(); i++) {
			i->second->read_sound_data(_sound, minsize);
		}
		dumped_size -= minsize;
#ifdef DEBUG_MIXER
		cout << "mixer::handle_frame() dumped_size == " << dumped_size << ", minsize == " << minsize << endl;
#endif
		_out->write(reinterpret_cast<unsigned short *>(_sound), minsize * 4);
	}
#ifdef DEBUG_MIXER
	cout << "mixer::handle_frame() : buffer completely field for this frame" << endl;
#endif
	return true;
}
