/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __MIXER_H_
#define __MIXER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef DEBUG
# ifndef NO_DEBUG_MIXER
#  define DEBUG_MIXER
# endif
#else
# ifdef DEBUG_MIXER
#  error DEBUG_MIXER defined without DEBUG
# endif
#endif

#include <exception>
#include <map>

class channel;

class sound_renderer;

/*! 	@brief The class for the player's sound mixer

	This class is used for sound mixing.
	It contains a list of current track and request them to mix.
	It then sends the mixed sound samples to the sound renderer.

*/
class mixer {
private:
	std::map<int, channel *> _channels;	//!< The map of track and channels
	int _frequency;					//!< The mixer's frequency
	short * _sound;					//!< Internal mixed sound buffer
	sound_renderer * _out;				//!< The sound renderer to send mixed sound data to.
public:
	mixer() throw();
	int get_frequency() const throw() { return _frequency; }
	virtual ~mixer() throw();
	/*!	@brief mixer initialisation
	
		This method initialize the sound mixer.
	
		@param out the sound renderer to send data to. must NOT be initialized, because it will be initialized by this method.
		@param freq the output frequency.
		@param size the recommended size of the internal sound buffer (unused)
	
		@return true if initialization was done successfully, false otherwise
	*/
	bool init(sound_renderer * out, int freq, int size) throw();
	/*!	@brief channel lookup
	
		This method allow to search for a particular channel, identified by it's track identifier.
	
		@param track the track identifier.
	
		@return a pointer if the channel exist, 0 otherwise
	*/
	channel * find_channel(int track) throw();
	/*!	@brief add a channel to the mixer
	
		This method allow to add a channel to the mixer.
	
		@param track the track identifier.
		@param c the channel to add.
	
		@return true if everything went fine, false otherwise
	*/
	bool add_channel(int track, channel * c) throw(std::exception);
	/*!	@brief performing mixing for a frame of the animation
	
		This method request that the mixer generates the sound data for the current frame and send it to the sound renderer.
	
		@return true if everything went fine, false otherwise
	*/
	bool handle_frame() throw(std::exception);
	/*!	@brief stop the mixer	
	
		@return true if everything went fine, false otherwise
	*/
	bool stop() throw(std::exception);
};

#endif
