/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#include <sstream>
#include <fstream>
#include <stdexcept>

#include "player.h"

#include "renderer.h"
#include "channel.h"
#include "chunck_type.h"
#include "rect.h"
#include "blitter.h"
#include "endian.h"

#ifdef DEBUG
#include <iomanip>
#include <iostream>
#endif

const int WAIT = 100;

using namespace std;

/*!	@brief parser and map of string resources

	This class implements a parser for the string resource format of SMUSH animations.
	It then allows the player to get the string corresponding to a particular identifier.
	
	@bug some of The Dig strings are not completely parsed (in titles)
*/
class string_resource {
private:
	map<int, string> _strings;	//!< The map of string resources (the key is the resource identifier)
public:
	/*!	@brief parse the given buffer

		@param buffer the buffer that contain the resource (in lucasart format)
		@param length the length of the buffer

		@return \c true if the parsing went fine, \c false otherwise
	*/
	bool init(char * buffer, int length) throw(exception) {
		string str(buffer, length);
		string::size_type i = str.find('#');
		while(i != string::npos) {
			string::size_type j = str.find('\n', i) + 1;
			assert(j != string::npos);
			string::size_type m = j - 2;
			while((str[m] == '\r') || (str[m] == '\n') || (str[m] == ' ') || (str[m] == '\t')) m--;
			//~ cout << static_cast<int>(str[m]) << " == " << str[m] << endl;
			string::size_type k = str.rfind(' ', m);
			string::size_type l = str.rfind('\t', m);
			if(l != string::npos && k < l) k = l;
			assert(k != string::npos); 
			m++; k++;
			//~ cout << "k == " << k << ", m == " << m << ", str == " << string(str, k, m-k) << endl;
			istringstream iss(string(str, k, m-k));
			int id;
			iss >> id;
			k = str.find('\n', j) - 1;
			if(str[j] == '/') j += 2;
			string value = string(str, j, k-j);
#ifdef DEBUG
			cout << "Inserting " << id << " == \"" << value << "\"" << endl;
#endif				
			_strings.insert(make_pair(id, value));
			i = str.find('#', k);
		}
		return true;
	}
	/*!	@brief extract a string

		@param id the resource identifier

		@return the corresponding string.
	*/
	const string & get(int id) const throw(exception) {
		map<int, string>::const_iterator i = _strings.find(id);
		if(i == _strings.end()) 
			throw(runtime_error("invalid string id"));
		return i->second;
	}
};

void player::show(const std::string & p) {
	if(p == "subtitles")
		_subtitles = true;
	else if(p == "bgmusic")
		_bgmusic = true;
	else if(p == "voices")
		_voices = true;
	else {
		istringstream iss(p);
		int id = -1;
		iss >> id;
		if(id < 0 || id > 36) throw runtime_error("invalid parameter to show");
		_skips.insert(make_pair(id, true));
	}
}

void player::hide(const std::string & p) {
	if(p == "subtitles")
		_subtitles = false;
	else if(p == "bgmusic")
		_bgmusic = false;
	else if(p == "voices")
		_voices = false;
	else {
		istringstream iss(p);
		int id = -1;
		iss >> id;
		if(id < 0 || id > 36) throw runtime_error("invalid parameter to show");
		_skips.insert(make_pair(id, false));
	}
}


player::player(renderer * renderer, bool wait, bool sound) throw() : 
							_version(-1), 
							_secondary_version(0),
							_sound_frequency(0),
							_nbframes(0), 
							_renderer(renderer), 
							_strings(0),
							_frame_size(-1, -1), 
							_frame(0),
							_output_sound(sound),
							_wait(wait),
							_already_init(false),
							_codec37_called(false),
							_skipnext(false),
							_subtitles(true),
							_bgmusic(true),
							_voices(true)
{
	_fr[0] = _fr[1] = _fr[2] = _fr[3] = 0;
	assert(_renderer != 0);
}

player::~player() throw() { 
	clean(); 
}

void player::update_palette(void) { 
	_renderer->set_palette(_pal); 
}

void player::clean() throw() {
	if(_strings) delete _strings;
	if(_fr[0]) delete _fr[0];
	if(_fr[1]) delete _fr[1];
	if(_fr[2]) delete _fr[2];
	if(_fr[3]) delete _fr[3];
}

void player::checkBlock(const chunck & b, chunck::type type_expected, unsigned int min_size) {
#ifdef DEBUG
	//~ cout << "checkBlock(" << chunck::chunck_string(b.get_type()) << ", " << b.get_size() << ")" << endl;
#endif		
	if(type_expected != b.get_type()) {
		ostringstream oss;
		oss << "chunck type is different from expected : " << b.get_type() << " != " << type_expected;
		throw runtime_error(oss.str());
	}
	if(min_size > b.get_size()) {
		ostringstream oss;
		oss << "chunck size is inferior than minimum required size : " << b.get_size() << " < " << min_size;
		throw runtime_error(oss.str());
	}
}

void player::handleSoundBuffer(int track_id, int index, int max_frames, int flags, int vol, int bal, chunck & b, int size) throw(exception) {
	if(!_voices && (flags & 128) == 128) return;
	if(!_bgmusic && (flags & 64) == 64) return;
	channel * c = _mixer.find_channel(track_id);
	if(c == 0) {
		c = new saud_channel(track_id,  _mixer.get_frequency());
		_mixer.add_channel(track_id, c);
	}
	if(index == 0)
		c->set_parameters(max_frames, flags, vol, bal);
	else
		c->check_parameters(index, max_frames, flags, vol, bal);
	c->append_data(b, size);
}

void player::handleSoundFrame(chunck & b) throw(exception) { 
#ifdef DEBUG
	cout << "handleSoundFrame()" << endl; 
#endif
	checkBlock(b, TYPE_PSAD);
	if(!_output_sound) return;
	int track_id = b.get_word();
	int index = b.get_word();
	int max_frames = b.get_word();
	int flags = b.get_word();
	int vol = b.get_byte();
	int bal = b.get_char();
#ifdef DEBUG
	//~ cout << "track_id == " << track_id << ", index == " << index << ", max_frames == " << max_frames << ", " << flags << ", " << vol << ", " << bal << endl;
	if(index == 0) { // What is thig flag about ??
		cout << "track_id == " << track_id << ", max_frames == " << max_frames << ", " << flags << ", " << vol << ", " << bal << endl;
	}
#endif
	int size = b.get_size() - 10;
	handleSoundBuffer(track_id, index, max_frames, flags, vol, bal, b, size);
}

void player::handleSkip(chunck & b) throw(exception) { 
	checkBlock(b, TYPE_SKIP, 4);
	int code = b.get_dword();
#ifdef DEBUG
	cout << "handleSkip(" << code << ")" << endl;
#endif
	map<int, bool>::iterator i = _skips.find(code);
	if(i != _skips.end())
		_skipnext = i->second;
	else
		_skipnext =true;
}

void player::handleStore(chunck & b) throw(exception) { 
#ifdef DEBUG
	cerr << "handleStore()" << endl; 
	//~ int i = 0;
	//~ while(!b.eof()) {
		//~ int code = b.get_byte();
		//~ cout << "STOR DATA " << i << " == " << code << endl;
		//~ i++;
	//~ }
#endif
	checkBlock(b, TYPE_STOR, 4);
	
}

void player::handleFetch(chunck & b) throw(exception) { 
#ifdef DEBUG
	cerr << "handleFetch()" << endl; 
	//~ int i = 0;
	//~ while(!b.eof()) {
		//~ int code = b.get_byte();
		//~ cout << "FTCH DATA " << i << " == " << code << endl;
		//~ i++;
	//~ }
#endif
	checkBlock(b, TYPE_FTCH, 6); 
}

void player::handleImuseBuffer(int track_id, int index, int nbframes, int size, int unk1, int unk2, chunck & b, int bsize) throw(exception) { 
	channel * c = _mixer.find_channel(track_id);
	if(c == 0) {
		c = new imuse_channel(track_id, _mixer.get_frequency());
		_mixer.add_channel(track_id, c);
	}
	if(index == 0)
		c->set_parameters(nbframes, size, unk1, unk2);
	else
		c->check_parameters(index, nbframes, size, unk1, unk2);
	c->append_data(b, bsize);
}

void player::handleImuseAction8(chunck & b, int flags, int unknown, int track_id) throw(exception) {
	assert(flags == 46 && unknown == 0);
	int unknown2 = b.get_word();
	track_id |= unknown2 << 16;
	int index = b.get_word();
	int nbframes = b.get_word();
	int size = b.get_dword();
#ifdef DEBUG
	cout << "handleImuseAction8() : " << unknown2 << ", " << index << "/" << nbframes << " == " << size << endl; 
#endif
	int bsize = b.get_size() - 18;
	handleImuseBuffer(track_id, index, nbframes, size, unknown, unknown2, b, bsize);
}

void player::handleImuseAction(chunck & b) throw(exception) { 
	checkBlock(b, TYPE_IACT, 8);
	if(!_output_sound) return;
	int code = b.get_word();
	int flags = b.get_word();
	int unknown = b.get_short();
	int track_id = b.get_word();
#ifdef DEBUG
	cout << "handleImuseAction(" << code << "," << flags << "," << unknown << "," << track_id << ")" << endl;
#endif
	switch(code) {
		case 8:
			handleImuseAction8(b, flags, unknown, track_id);
			break;
#ifdef DEBUG
		default:
			//~ cerr << "unknown IACT code : " << code << ", IACT size == " << b.get_size() << endl;
			//~ cout << "unknown IACT code : " << code << ", IACT size == " << b.get_size() << endl;
			{
				cerr  << setw(5) << track_id << " " << code << " " << setw(8) << flags << " " << setw(4) << unknown << " : ";
				int i = 1;
				while(!b.eof()) {
					int value = b.get_byte();
					cerr << setw(3) << value << " ";
					if(i++ == 80) { i = 0; cerr << endl; }
				}
				if(i) cerr << endl;
			}
#endif
	}
}

//~ #include <iostream>

void player::handleTextResource(chunck & b) throw(exception) { 
#ifdef DEBUG
	cout << "handleTextResource()" << endl; 
#endif
	checkBlock(b, TYPE_TRES, 18); 
	assert(_strings);
	int pos_x = b.get_short();
	int pos_y = b.get_short();
	int flags = b.get_short();
	int left = b.get_short();
	int top = b.get_short();
	int width = b.get_short();
	int height = b.get_short();
	int unk2 = b.get_word();
	int string_id = b.get_word();
	string str = _strings->get(string_id);
#ifdef DEBUG
	//~ cerr << "TRES : " << _subtitles << ", " <<  flags << ", " << left << ", " << top << ", " << width << ", " << height << ", " << unk2 << " : " << str << endl;
#else
	height = height;
	unk2 = unk2;
#endif
	if((!_subtitles) && ((flags & 8) == 8)) return;
	font_renderer * fr = _fr[0];
	int color = 15;
	while(str[0] == '^') {
		switch(str[1]) {
		case 'f':
			{
				istringstream iss(string(str, 2, 2));
				int id = 0;
				iss >> id;
				str = string(str, 4); 
				fr = _fr[id]; 
			} break;
		case 'c':
			{
				istringstream iss(string(str, 2, 3));
				iss >> color;
				str = string(str, 5); 
			} break;
		default:
			throw runtime_error("invalid escape code in text string");
		}
	}
	fr->set_color(color);
	// if subtitles disabled and bit 3 is set, then do not draw
	if(!_cur_buffer) { _cur_buffer = _renderer->lock_frame(_frame); }
	if(flags == 0 || flags == 4)
		fr->draw_string_absolute(str, _cur_buffer, _frame_size, pos_x, pos_y);
	else 
		fr->draw_string_centered(str, _cur_buffer, _frame_size, max(pos_y, top), left, width, pos_x);
}

void player::readPalette(palette & out, chunck & in) throw(exception) {
	unsigned char buffer[768];
	in.read(buffer, 768);
	out = palette(buffer);
}

void player::handleDeltaPalette(chunck & b) throw(exception) { 
#ifdef DEBUG_DELTA_PALETTE
	cout << "handleDeltaPalette()" << endl; 
#endif
	checkBlock(b, TYPE_XPAL);
	if(b.get_size() == 768*3+4) {
		int unk1, num;
		unk1 = b.get_word();
		num = b.get_word();
		for(int i = 0; i < 768; i++)
		{
			_delta_pal[i] = b.get_word();
		}
		readPalette(_pal, b);
		update_palette();
#ifdef DEBUG_DELTA_PALETTE
		cout << "DELTA PALETTE : unk1 == " << unk1 << ", num == " << num << endl;
		for(int i = 0; i < 256; i++)
		{
			cout << "deltapal[" << i << "] == " << _delta_pal[3*i] << ", " << _delta_pal[3*i+1] << ", " << _delta_pal[3*i+2] << endl;
		}
		cout << "palette == " << _pal << endl;
#endif
	} else if(b.get_size() == 6) {
		int unk1, num, unk2;
		unk1 = b.get_word();
		num = b.get_word();
		unk2 = b.get_word();
#ifdef DEBUG_DELTA_PALETTE
		cout << "DELTA PALETTE : unk1 == " << unk1 << ", num == " << num << ", unk2 == " << unk2 << endl;
#endif
		for(int i = 0; i < 256; i++)
		{
			_pal[i].delta(_delta_pal+3*i);
		}
		update_palette();
	} else {
		throw runtime_error("wrong size for DeltaPalette");
	}
}

void player::handleNewPalette(chunck & b) throw(exception) { 
#ifdef DEBUG
	cout << "handleNewPalette()" << endl; 
#endif
	checkBlock(b, TYPE_NPAL, 768);
	readPalette(_pal, b);
	update_palette();
}

void player::decodeCodec(chunck & b, const rect & r, decoder & codec) throw(exception) {
	assert(_cur_buffer);
	try {
		blitter blit(_cur_buffer, _frame_size, r);
		codec.decode(blit, b);
	} catch(const exception & e) {
#ifdef DEBUG
		cerr << "Exception occured during decodeCodec : " << e.what() << endl;
#endif
	}
}

void player::init_size(const rect & r, bool always, bool transparent) throw(exception)
{
	if(_codec37_called) _already_init = true;
		
	if(!_already_init || _frame_size.get_x() < r.right() || _frame_size.get_y() < r.bottom() || always) {
		if(_cur_buffer) { _renderer->unlock_frame(); _cur_buffer = 0; }
		_frame_size = r.bottomright();
		_renderer->init_frame(_frame_size);
	}
	if(_cur_buffer) { _renderer->unlock_frame(); _cur_buffer = 0; }
	_cur_buffer = _renderer->lock_frame(_frame);
	if(!_already_init && transparent) {
#ifdef DEBUG
		cout << "clearing buffer" << endl;
#endif	
		memset(_cur_buffer, 0, _frame_size.get_x()*_frame_size.get_y());
	}
	_codec1.init_size(_frame_size, r);
	_codec37.init_size(_frame_size, r);
	_codec44.init_size(_frame_size, r);
	_already_init = true;
}

void player::handleFrameObject(chunck & b) throw(exception) { 
#ifdef DEBUG
	cout << "handleFrameObject()" << endl; 
#endif
	checkBlock(b, TYPE_FOBJ, 14);
	if(_skipnext) {
		_skipnext = false;
		return;
	}
	int codec = b.get_word();
	unsigned short left = b.get_word();
	unsigned short top = b.get_word();
	unsigned short width = b.get_word();
	unsigned short height = b.get_word();
	rect r(left, top, left+width, top+height);
	unsigned short data[2];
	data[1] = b.get_word();
	data[0] = b.get_word();
#ifdef DEBUG
	cout << "Frame pos : " << left << ", " << top << endl;
	cout << "Frame size : " << width << "x" << height << endl;
	cout << "Codec : " << codec << endl;
	for(int i = 0; i < 2; i++)
		if(data[i] != 0) {
			cout << "FRAMEOBJECT:data[" << i << "] == " << data[i] << endl;
			cerr << "FRAMEOBJECT:data[" << i << "] == " << data[i] << endl;
		}
#endif
	switch (codec) {
	case 3:
	case 1:
		init_size(r, false, true);
		decodeCodec(b, r, _codec1);
		break;
	case 37:
		assert(left == 0 && top == 0);
		init_size(r, true, false);
		decodeCodec(b, r, _codec37);
		_codec37_called = true;
		break;
	case 21:
	case 44:
		init_size(r, true, true);
		decodeCodec(b, r, _codec44);
		break;
	default:
		ostringstream oss;
		oss << "Invalid codec for frame object : " << static_cast<int>(codec);
		throw runtime_error(oss.str());
	}
	//~ _renderer->save_current();
}


void player::handleFrame(chunck & b) throw(exception) {
#ifdef DEBUG
	cout << "handleFrame(" << _frame << ")" << endl; 
#endif
	checkBlock(b, TYPE_FRME);
	_already_init = false;
	_skipnext = false;
	
	while(!b.eof()) {
		chunck * sub = b.sub_block();
		if(sub->get_size() & 1) b.seek(1);
		switch(sub->get_type()) {
			case TYPE_NPAL:
				handleNewPalette(*sub);
				break;
			case TYPE_FOBJ:
				handleFrameObject(*sub);
				break;
			case TYPE_PSAD:
				handleSoundFrame(*sub);
				break;
			case TYPE_TRES:
				handleTextResource(*sub);
				break;
			case TYPE_XPAL:
				handleDeltaPalette(*sub);
				break;
			case TYPE_IACT:
				handleImuseAction(*sub);
				break;
			case TYPE_STOR:
				handleStore(*sub);
				break;
			case TYPE_FTCH:
				handleFetch(*sub);
				break;
			case TYPE_SKIP:
				handleSkip(*sub);
				break;
			default:
				ostringstream oss;
				oss << "Unknown frame subchunck found : " << chunck::chunck_string(sub->get_type()) << ", " << sub->get_size();
				throw runtime_error(oss.str());
		}
		delete sub;
	}
	if(_cur_buffer) { _renderer->unlock_frame(); _cur_buffer = 0; }
	if(_output_sound) _mixer.handle_frame();
#ifdef DEBUG
	//~ cout << "Frame() terminated" << endl; 
	cout << "=========================================================" << endl; 
#endif
	_renderer->flip_frame();
	if(_wait) _renderer->wait(WAIT);
	_frame++;
}

void player::handleAnimHeader(chunck & b) throw(exception) {
#ifdef DEBUG
	cout << "handleAnimHeader()" << endl; 
#endif
	checkBlock(b, TYPE_AHDR, 774); 
	_version = b.get_word();
	_nbframes = b.get_word();
	int unknown = b.get_word();
#ifdef DEBUG
	cout << "HEADER : version == " << _version << ", nbframes == " << _nbframes << ", unknown == " << unknown << endl;
#else
	unknown = unknown;
#endif
	_renderer->start_decode(_fname, _version, _nbframes);
	readPalette(_pal, b);
	update_palette();
	if(_version == 1) {
		_sound_frequency = 22050;
	}
	if(_version == 2) {
		_secondary_version = b.get_dword();
		int unknown2 = b.get_dword();
		_sound_frequency = b.get_dword();
#ifdef DEBUG
		cout << "HEADER : secondary version == " << _secondary_version << ", unknown2 == " << unknown2 << ", sound frequency == " << _sound_frequency << endl;
		int i = 0, c;
		while(!b.eof()) {
			c = b.get_byte();
			//~ cout << "HEADER : remaining bytes : " << i << " == " << c << endl;
			if(c) cerr << "HEADER : remaining bytes : " << i << " == " << c << endl;
			i++;
		}
#else
		unknown2 = unknown2;
#endif
		if(_secondary_version != 10 && _secondary_version != 0 && _secondary_version != 12 && _secondary_version != 15 && _secondary_version != 14)
			throw runtime_error("Wrong secondary version number for SMUSH animation");
		if(_sound_frequency != 0 && _sound_frequency != 11025 && _sound_frequency != 22050)
			throw runtime_error("Wrong _sound_frequency number for SMUSH animation");
	}else if(_version > 2) {
		throw runtime_error("Wrong primary version number for SMUSH animation");
	}
	if(_output_sound && _sound_frequency) {
		if(_sound_frequency != 22050) _sound_frequency = 22050;
		sound_renderer * sr = _renderer->get_sound_renderer();
		if(sr) {
			_mixer.init(sr, _sound_frequency, 1000 / WAIT);
		} else {
			_output_sound = false;
		}
	}
}

string_resource * get_strings(const string & file, bool is_encoded) {
	ifstream is;
	is.exceptions(ifstream::badbit);
	is.open (file.c_str(), ios::binary);
	if(is.fail()) return 0;
	is.seekg (0, ios::end);
	ifstream::pos_type length = is.tellg();
	is.seekg (0, ios::beg);
	char * filebuffer = new char [length];
	assert(filebuffer);
	is.read (filebuffer,length);
	is.close();
	if(is_encoded) {
		chunck::type type = from_big_endian(*reinterpret_cast<unsigned int *>(filebuffer));
		if(type != TYPE_ETRS) throw runtime_error("invalid type for file"); // mem leak !!!
		char * old = filebuffer;
		filebuffer = new char[static_cast<int>(length) - 8];
		for(int i = 8; i < length; i++)
			filebuffer[i-8] = old[i] ^ 0xCC;
		delete []old;
		length -= 8;
	}
	string_resource * sr = new string_resource;
	assert(sr);
	try {
		sr->init(filebuffer, length);
	} catch(exception & e) {
		delete sr;
		sr = 0;
	}
	delete []filebuffer;
	return sr;
}

bool player::read_string(const string & file) throw(exception) {
#ifdef DEBUG
	cout << "player::read_string(" << file << ")" << endl;
#endif
	string::size_type i = file.rfind('.');
	if(i == string::npos) throw(runtime_error("invalid filename"));
	string fname = string(file, 0, i) + ".TRS";
	if((_strings = get_strings(fname, false)) != 0)
		return true;
	i = file.rfind('\\');
	if(i == string::npos) i = file.rfind('/');
	if(i == string::npos) throw(runtime_error("invalid filename"));
	fname = string(file, 0, i+1) + "MINEROAD.TRS";
	if((_strings = get_strings(fname, false)) != 0)
		return true;
	fname = string(file, 0, i+1) + "DIGTXT.TRS";
	if((_strings = get_strings(fname, true)) != 0)
		return false;
	return true;
}

namespace {
	font_renderer * load_font(const string & file, bool original = false) {
#ifdef DEBUG
		cout << "loading font from \"" << file << "\"" << endl;
#endif
		font_renderer * fr = new font_renderer(original);
		player p(fr, false, false);
		p.play(file.c_str());
		return fr;
	}
}

bool player::play(const string & file) throw(exception) {
	string::size_type i = file.rfind('\\');
	if(i == string::npos)
	{
		i = file.rfind('/');
	} else {
		string::size_type j = file.rfind('/', i);
		if(j != string::npos)
			i = j;
	}
	string directory;
	if(i != string::npos) {
		directory = string(file, 0, i);
		_fname = string(file, i+1);
	} else {
		directory = "";
		_fname = string(file);
	}
	clean();
	
	bool is_fullthrottle = true;
	try {
		is_fullthrottle = read_string(file);
	} catch(exception & e) {
#ifdef DEBUG
		cerr << "read_string() returned an exception : " << e.what();
#endif
	}
	if(_strings) {
		if(is_fullthrottle) {
			if(directory == "") {
				directory == "../DATA/";
			} else {
				string::size_type i = directory.rfind('\\');
				if(i == string::npos) i = directory.rfind('/');
				if(i == string::npos) {
					directory = "DATA/";
				} else {
					directory = string(directory, 0, i) + "/DATA/";
				}
			}
			_fr[0] = load_font(directory + "SCUMMFNT.NUT", true);
			_fr[2] = load_font(directory + "TITLFNT.NUT", true);
		} else {
			for(int i = 0; i < 4; i++) {
				ostringstream oss;
				oss << directory << "/FONT" << i << ".NUT";
				_fr[i] = load_font(oss.str());
			}
		}
	}
	
	file_chunck base = file_chunck(file);
	
	checkBlock(base, TYPE_ANIM); 
	
	while(!base.eof()) {
		chunck * sub = base.sub_block();
		switch(sub->get_type()) {
			case TYPE_AHDR:
				handleAnimHeader(*sub);
				break;
			case TYPE_FRME:
				handleFrame(*sub);
				break;
			default:
				ostringstream oss;
				oss << "Unknown chunck found : " << sub->get_type() << ", " << sub->get_size();
				throw runtime_error(oss.str());
		}
		delete sub;
	}
	if(_output_sound) {
		_mixer.stop();
	}
	return true;
}
