/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __PLAYER_H_
#define __PLAYER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <exception>
#include <string>
#include <map>

#include "rect.h"
#include "mixer.h"
#include "chunck.h"
#include "palette.h"
#include "codec1.h"
#include "codec37.h"
#include "codec44.h"
#include "frenderer.h"

class renderer;
	
class string_resource;

/*!	@brief the SMUSH player class

	This class is the player itself. 
*/
class player {
private:
	std::string _fname;		//!< the name of the animation file being played
	int _version;			//!< the version of the animation file being played
	int _secondary_version;	//!< the secondary version number of the animation file being played
	int _sound_frequency;		//!< the sound frequency of the animation file being played
	int _nbframes;			//!< the number of frames in the animation file
	mixer _mixer;			//!< the sound mixer
	palette _pal;			//!< the current palette
	short _delta_pal[768];		//!< the delta palette information set by an xpal
	renderer * _renderer;		//!< pointer to the ::renderer
	string_resource * _strings;	//!< pointer to the string resources associated with the animation
	font_renderer * _fr[4];		//!< pointers to the fonts for the animation
	codec1_decoder _codec1;	//!< the ::decoder for codec 1 and 3
	codec37_decoder _codec37;	//!< the ::decoder for codec 37
	codec44_decoder _codec44;	//!< the ::decoder for codec 21 and 44
	point _frame_size;		//!< the current frame size of the animation
	int _frame;				//!< the current frame number of the animation
	bool _output_sound;		//!< should we handle sound ?
	bool _wait;				//!< should we synchronise the player ?
	bool _already_init;		//!< has the player already been initialized for the current frame
	bool _codec37_called;		//!< has the codec 37 already been called once for this animation
	bool _skipnext;			//!< should the player skip the next frame object ?
	bool _subtitles;			//!< should the player handle subtitles ?
	bool _bgmusic;			//!< should the player output the background music ?
	bool _voices;			//!< should the player output the voice ?
	std::map<int, bool> _skips;	//!< mapping of frame object identifier to show or hide
	char * _cur_buffer;		//!< pointer to the current frame
public:
	player(renderer *, bool wait = true, bool output_sound = true) throw();
	virtual ~player() throw();
	bool play(const std::string &) throw(std::exception);
	void update_palette(void);
	void show(const std::string & p);
	void hide(const std::string & p);
protected:
	bool read_string(const std::string & file) throw(std::exception);
	void clean() throw();
	void checkBlock(const chunck &, chunck::type, unsigned int = 0);
	void handleAnimHeader(chunck &) throw(std::exception);
	void handleFrame(chunck &) throw(std::exception);
	void handleNewPalette(chunck &) throw(std::exception);
	void handleFrameObject(chunck &) throw(std::exception);
	void handleSoundBuffer(int, int, int, int, int, int, chunck &, int) throw(std::exception);
	void handleImuseBuffer(int, int, int, int, int, int, chunck &, int) throw(std::exception);
	void handleSoundFrame(chunck &) throw(std::exception);
	void handleSkip(chunck &) throw(std::exception);
	void handleStore(chunck &) throw(std::exception);
	void handleFetch(chunck &) throw(std::exception);
	void handleImuseAction8(chunck &, int flags, int unknown, int track_id) throw(std::exception);
	void handleImuseAction(chunck &) throw(std::exception);
	void handleTextResource(chunck &) throw(std::exception);
	void handleDeltaPalette(chunck &) throw(std::exception);
	void decodeCodec(chunck &, const rect &, decoder &) throw(std::exception);
	void readPalette(palette &, chunck &) throw(std::exception);
	void init_size(const rect &, bool, bool) throw(std::exception);
};

#endif
