/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __RECT_H_
#define __RECT_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ostream>

/*! 	@brief simple class for handling both 2D position and size

	This small class is an helper for position and size values.
*/
class point {
private:
	int _x;	//!< The horizontal part of the point
	int _y;	//!< The vertical part of the point
public:
	point() throw() : _x(0), _y(0) {};
	point(const point & p) throw() : _x(p.get_x()), _y(p.get_y()) {};
	explicit point(int x, int y) throw() : _x(x), _y(y) {};
	point & operator=(const point & p) throw() { _x = p.get_x(); _y = p.get_y(); return *this; };
	bool operator==(const point & p) const throw() { return _x == p.get_x() && _y == p.get_y(); };
	const int & get_x() const throw() { return _x; };
	const int & get_y() const throw() { return _y; };
	int & get_x() throw() { return _x; };
	int & get_y() throw() { return _y; };
	point operator+(const point & p) const throw() { return point(_x+p.get_x(), _y+p.get_y()); };
	point operator-(const point & p) const throw() { return point(_x-p.get_x(), _y-p.get_y()); };
	point & operator+=(const point & p) throw() { _x += p.get_x(); _y += p.get_y(); return *this; };
	point & operator-=(const point & p) throw() { _x -= p.get_x(); _y -= p.get_y(); return *this; };
	bool is_origin() const throw() { return *this == point(0, 0); };
	void set(int x, int y) throw() { _x = x; _y = y; }
};

/*! 	@brief simple class for handling a rectangular zone.

	This small class is an helper for rectangles.
	It is mostly used by the blitter class.
*/
class rect {
private:
	point _topleft;		//!< The point at the top left of the rectangle
	point _bottomright;	//!< The point at the bottom right of the rectangle
protected:
	void check() throw(std::exception);
public:
	rect() throw();
	rect(int x, int y) throw(std::exception);
	explicit rect(const point & size) throw(std::exception);
	rect(int x1, int y1, int x2, int y2) throw(std::exception);
	rect(const point & topleft, const point & bottomright) throw(std::exception);
	rect(const rect & r) throw();
	rect & operator=(const rect & r) throw();
	bool operator==(const rect & r) const throw();
	point size() const throw() { return (_bottomright - _topleft); };
	int width() const throw() { return size().get_x(); }
	int height() const throw() { return size().get_y(); }
	int left() const throw() { return _topleft.get_x(); }
	int right() const throw() { return _bottomright.get_x(); }
	int top() const throw() { return _topleft.get_y(); }
	int bottom() const throw() { return _bottomright.get_y(); }
	const point & topleft() const throw() { return _topleft; }
	const point & bottomright() const throw() { return _bottomright; }
	
	/*!	@brief check if given position is inside the rectangle
		
		@param x the horizontal position to check
		@param y the vertical position to check	
		
		@return true if the given position is inside the rectangle, false otherwise
	*/
	bool is_inside(int x, int y) const throw();
	/*!	@brief check if given point is inside the rectangle
		
		@param p the point to check
		
		@return true if the given point is inside the rectangle, false otherwise
	*/
	bool is_inside(const point & p) const throw();
	bool clip(rect & r) const throw(std::exception);
};

extern std::ostream & operator<<(std::ostream & o, const rect & r);
extern std::ostream & operator<<(std::ostream & o, const point & p);

#endif
