/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __RENDERER_H_
#define __RENDERER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "rect.h"

#include <string>
#include <exception>

class palette;
	
/*! 	@brief interface for sound output

	This is the interface for sound output. Only one implementation is available right now : ds_wave_out
*/
class sound_renderer {
public:
	virtual ~sound_renderer() {};
	/*!	@brief sound output initialisation
		
		This is called by the mixer when initialisation has to be performed.
		
		@param frequency is always 22050 right now.
		@param stereo is true if the sound output should be stereo (always true right now)

		@return true if initialisation was ok, false otherwise
	*/
	virtual bool init(int frequency, bool stereo) = 0;
	/*!	@brief sound data output
		
		This is called by the mixer to put data to the output.
		
		@param ptr buffer containing samples.
		@param size size of the \c ptr buffer, in samples (i.e. size in byte / 2)

		@return true if everything went fine, false otherwise
	*/
	virtual bool write(unsigned short * ptr, int size) = 0;
	/*!	@brief stop sound output
		
		This is called by the mixer when sound output should be stopped.

		@return true if everything went fine, false otherwise
	*/
	virtual bool stop() = 0;
};

/*! 	@brief interface for general output (rendering)

	This is the interface for frame output. 
	Several implementations of these interface exist, each having a particular
	application.
*/
class renderer {
public:
	virtual ~renderer() {};
	/*!	@brief start of animation output
		
		This is called by the animation player when output is going to start.
		
		@param fname name of the animation being played.
		@param version version number of the animation
		@param nbframes total number of frames of the animation.
		
		@return true if initialisation was ok, false otherwise
	*/
	virtual bool start_decode(const std::string & fname, int version, int nbframes) throw(std::exception) = 0;
	/*!	@brief start of animation output
		
		This is called by the animation player when the frame size is changing.
		
		@param size new size of the frames.
		
		@return true if everything went fine, false otherwise
	*/
	virtual bool init_frame(const point & size) throw(std::exception) = 0;
	/*!	@brief set a new palette
		
		This is called by the animation player when the palette is changing.
		
		@param pal new palette.
		
		@return true if everything went fine, false otherwise
	*/
	virtual bool set_palette(const palette & pal) throw(std::exception) = 0;
	/*!	@brief lock a frame buffer
		
		This is called by the animation player when a frame is going to be decoded.
		
		@param frame the frame number.
		
		@return a pointer to the frame buffer to output data to.
	*/
	virtual char * lock_frame(int frame) throw(std::exception) = 0;
	/*!	@brief unlock a frame buffer
		
		This is called by the animation player when a frame has been decoded.
		
		@return true if everything went fine, false otherwise
	*/
	virtual bool unlock_frame() throw(std::exception) = 0;
	/*!	@brief flip a frame buffer
		
		This is called by the animation player when the current frame should be shown.
		
		@return true if everything went fine, false otherwise
	*/
	virtual bool flip_frame() throw(std::exception) = 0;
	/*!	@brief wait for some time
		
		This is called by the animation player when the animation should stay idle.
		
		@param ms number of millisecond to wait.
		
		@return true if everything went fine, false otherwise
	*/
	virtual bool wait(int ms) throw(std::exception) = 0;
	/*!	@brief request for a sound_renderer
		
		This is called by the animation player when sound output is required by the animation.
		
		@return a valid pointer to an uninitialized sound_renderer instance, or null if none is available.
	*/
	virtual sound_renderer * get_sound_renderer() throw(std::exception) = 0;
	/*!	@brief debugging function : do not use
		
		@return true if everything went fine, false otherwise
	*/
	virtual bool save_current() throw() { return false; };
};

#endif
