/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#include "win32.h"
#include <sstream>
#include <iomanip>
#include <fstream>
#include <stdexcept>
#include <cstring>
#include <mmsystem.h>
#include <dsound.h>
#ifdef DEBUG_WIN32
#include <iostream>
#endif


using namespace std;


const GUID GUID_NULL = {0};
/*!	@brief DirectSound 8 sound renderer

	This class implements a DirectSound8 compatible sound renderer.

	\todo redo the synchronisation layer.
	\bug The synchronisation takes 100% of CPU cycles.
	\bug The synchronisation may fail on unpredictable occasion (the time frame when sync occurs is too small).
*/
class ds_wave_out : public sound_renderer {
private:
	IDirectSound8 * _ds;
	IDirectSoundBuffer8 * _dsb;
	int _writepos;
	int _oldwritepos;
	//~ int _totalwrite;
	//~ int _totalwritecall;
	int _size;
	bool _started;
public:
	ds_wave_out() : _ds(0), _dsb(0), _started(false) {};
	virtual ~ds_wave_out() {
		if(_dsb) { _dsb->Release(); _dsb = 0; }
		if(_ds) { _ds->Release(); _ds = 0; }
		//~ if(_totalwritecall)
			//~ cerr << "DirectX Sound playback : _totalwrite == " << _totalwrite << ", _totalwritecall == " << _totalwritecall << ", " << ((double)_totalwrite) / _totalwritecall << endl;
	}
	bool init(int frequency, bool stereo) {
		HRESULT hr = DirectSoundCreate8(&DSDEVID_DefaultPlayback, &_ds, NULL);
		if(FAILED(hr)) throw runtime_error("DirectSoundCreate8() failed");
		hr = _ds->SetCooperativeLevel(GetForegroundWindow(), DSSCL_NORMAL);
		if(FAILED(hr)) throw runtime_error("SetCooperativeLevel() failed");
		WAVEFORMATEX wf;
		wf.wBitsPerSample = 16;
		wf.nChannels = stereo ? 2 : 1;
		wf.nSamplesPerSec = frequency;
		wf.nBlockAlign = wf.nChannels * wf.wBitsPerSample / 8;
		wf.wFormatTag = WAVE_FORMAT_PCM;
		wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign;
		wf.cbSize =0;
		
		_size =  wf.nAvgBytesPerSec * 5;
		_writepos = 0;
		_oldwritepos = 0;
		//~ _totalwrite = 0;
		//~ _totalwritecall = 0;
		DSBUFFERDESC desc;
		desc.dwSize = sizeof(desc);
		desc.dwFlags = DSBCAPS_GETCURRENTPOSITION2|DSBCAPS_GLOBALFOCUS|DSBCAPS_STICKYFOCUS;
		desc.dwBufferBytes = _size; // 5 second of playback...
		desc.dwReserved = 0;
		desc.lpwfxFormat = &wf;
		desc.guid3DAlgorithm = GUID_NULL;
		hr = _ds->CreateSoundBuffer(&desc, (IDirectSoundBuffer**)&_dsb, NULL);
		if(FAILED(hr)) throw runtime_error("CreateSoundBuffer() failed");
		
		return true;
	}
	bool write(unsigned short * ptr, int size) {
		assert(_dsb && _ds);
		assert(size && (size < _size) && ((size % 4) == 0));

		unsigned short * ptr1, *ptr2;
		DWORD size1, size2;
		//~ _totalwritecall++;
		//~ _totalwrite += size;
		//~ size *= sizeof(unsigned short);
		_dsb->GetCurrentPosition(&size1, &size2);
#ifdef DEBUG_WIN32
		cout << "ds_wave_out::write() == " << _writepos << ", " << size << ", " << _writepos + size << endl;
		cout << "DirectSoundBuffer8->GetCurrentPosition() == " << size1 << ", " << size2 << endl;
#endif
		HRESULT hr = _dsb->Lock(_writepos, size, (void**)&ptr1, &size1, (void**)&ptr2, &size2, 0);
		if(SUCCEEDED(hr)) {
			memcpy(ptr1, ptr, size1);
			if(size2) memcpy(ptr2, reinterpret_cast<char*>(ptr)+size1, size2);
			_dsb->Unlock((void*)ptr1, size1, (void*)ptr2, size2);
		} else {
			ostringstream oss;
			oss << "_dsb->Lock() failed : 0x" << hex << hr  << " == " << DSERR_INVALIDPARAM << endl;
			throw runtime_error(oss.str());
		}
		if(!_started) { 
#ifdef DEBUG_WIN32
			cout << "Starting DirectX Sound playback" << endl;
#endif
			_dsb->Play(0, 0, DSBPLAY_LOOPING); 
			_started = true; 
		}
		_oldwritepos = _writepos;
		_writepos += size;
		if(_writepos >= _size) _writepos -= _size;
		
		return true;
	}
	bool wait() {
		assert(_dsb && _ds);
		if(!_started) return false;
		DWORD play;
		while(1) {
			HRESULT hr = _dsb->GetCurrentPosition(&play, NULL);
			if(FAILED(hr)) return false;
			if(_writepos < _oldwritepos && (play > _oldwritepos || play < _writepos)) return true;
			if(_writepos > _oldwritepos && play > _oldwritepos && play < _writepos) return true;
			//~ if(play >= _oldwritepos && play < _writepos) return true;
			//~ if(ret) cout << "ds_wave_out::wait() _writepos == " << _writepos << ", _oldwritepos == " << _oldwritepos << ", play == " << play << endl;
		}
	}
	bool stop() {
		const int SILENCE_SIZE = 2205*4; // 1 tenth of a second of silence, please...
		assert(_dsb && _ds);
		if(!_started) return true;
		// First we put some silence...
		unsigned short * ptr1, *ptr2;
		DWORD size1, size2;
		HRESULT hr = _dsb->Lock(_writepos, SILENCE_SIZE, (void**)&ptr1, &size1, (void**)&ptr2, &size2, 0);
		if(SUCCEEDED(hr)) {
			memset(ptr1, 0, size1);
			if(size2) memset(ptr2, 0, size2);
			_dsb->Unlock((void*)ptr1, size1, (void*)ptr2, size2);
		} else {
			throw runtime_error("_dsb->Lock() failed");
		}
//		_oldwritepos = _writepos;
//		_writepos = (_writepos + SILENCE_SIZE) % _size;
		//~ cerr << "_oldwritepos == " << _oldwritepos << ", _writepos == " << _writepos << endl;
		// Then we wait for the silence to be reached
		if(_writepos + SILENCE_SIZE > _size)
			_writepos = 0;
		while(1) {
			DWORD play;
			_dsb->GetCurrentPosition(&play, NULL);
			if(play > _writepos) {
				_dsb->Stop();
				_started = false;
				return true;
			}
		}
		return true;
	}
};


bitmap_renderer::bitmap_renderer() throw(exception) : _padding(0) {
	_bmi = reinterpret_cast<BITMAPINFO *>(new char[HEADER_SIZE]);
	if(!_bmi) throw runtime_error("could not allocate bitmap header");
	_bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	_bmi->bmiHeader.biWidth = 0;
	_bmi->bmiHeader.biHeight = 0;
	_bmi->bmiHeader.biPlanes = 1;
	_bmi->bmiHeader.biBitCount = 8;
	_bmi->bmiHeader.biCompression = BI_RGB;
	_bmi->bmiHeader.biSizeImage = 0;
	_bmi->bmiHeader.biXPelsPerMeter = 300;
	_bmi->bmiHeader.biYPelsPerMeter = 300;
	_bmi->bmiHeader.biClrUsed = 256;
	_bmi->bmiHeader.biClrImportant = 0;
	for(int i = 0; i < 256; i++)
	{
		_bmi->bmiColors[i].rgbBlue = 0;
		_bmi->bmiColors[i].rgbGreen = 0;
		_bmi->bmiColors[i].rgbRed = 0;
		_bmi->bmiColors[i].rgbReserved = 0;
	}
}

bitmap_renderer::~bitmap_renderer() throw() {
	delete[] reinterpret_cast<char *>(_bmi);
}

bool bitmap_renderer::init_frame(const point &p) throw(exception) {
	_bmi->bmiHeader.biWidth = p.get_x();
	_bmi->bmiHeader.biHeight = p.get_y();
	_padding = p.get_x();
	if(_padding & 3) {
		_padding = ((_padding >> 2) + 1) << 2;
	}
	_padding -= p.get_x();
	return base_renderer::init_frame(p);
}

bool bitmap_renderer::set_palette(const palette & p) throw() {
	base_renderer::set_palette(p);
	for(int i = 0; i < 256; i++)
	{
		_bmi->bmiColors[i].rgbRed = pal()[i].red();
		_bmi->bmiColors[i].rgbGreen = pal()[i].green();
		_bmi->bmiColors[i].rgbBlue = pal()[i].blue();
	}
	return true;
}

void bitmap_renderer::save(int frame) throw(exception) {
	int oldframe = get_frame();
	if(frame == -1) frame = get_frame();
	else set_frame(frame);
	//~ cerr << "bitmap_renderer::save(" << frame << ")" << endl;
	int real_width = get_width()+_padding;
	char * bitmap = new char[get_height() * real_width];
	if(!bitmap) throw runtime_error("could not allocate bitmap buffer");
	for(int i = 0; i < get_height(); i++) {
		memcpy(bitmap + i * real_width, data() + (get_height()-1-i)*get_width(), get_width());
	}
	dump(bitmap, get_height() * real_width);
	delete []bitmap;
	set_frame(oldframe);
}

bitmap_file_renderer::bitmap_file_renderer(const string & name) throw() : _name(name) {
	_bfi.bfType = 'MB';
	_bfi.bfSize = HEADER_SIZE + sizeof(_bfi);
	_bfi.bfReserved1 = 0;
	_bfi.bfReserved2 = 0;
	_bfi.bfOffBits = HEADER_SIZE + sizeof(_bfi);
}

bool bitmap_file_renderer::save_current() throw()
{
static int _current = 10000;
	save(_current++);
	return true;
}

bool bitmap_file_renderer::init_frame(const point &p) throw() {
	int padding = p.get_x();
	if(padding & 3) {
		padding = ((padding >> 2) + 1) << 2;
	}
	_bfi.bfSize = padding * p.get_y() + HEADER_SIZE + sizeof(_bfi);
	return bitmap_renderer::init_frame(p);
}

void bitmap_file_renderer::dump(char * data, int size) throw(exception) {
	ostringstream oss;
	oss << _name << setw(5) << setfill('0') << get_frame() << setw(0) << ".bmp";
	ofstream ofs;
	ofs.open(oss.str().c_str(), ios::binary);
	ofs.write(reinterpret_cast<const char *>(&_bfi), sizeof(_bfi));
	ofs.write(reinterpret_cast<const char *>(get_bmi()), HEADER_SIZE);
	ofs.write(data, size);
}

LRESULT WINAPI window_renderer::windowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg) {
	case WM_USER:
		DestroyWindow(hwnd);
		break;
	case WM_USER+1:
		{
			window_renderer * wr = reinterpret_cast<window_renderer *>(GetWindowLong(hwnd, 0));
			SetEvent(wr->_hEvent);
		} break;
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	case WM_CREATE:
		{
			LPCREATESTRUCT lpcs = reinterpret_cast<LPCREATESTRUCT>(lParam);
			SetWindowLong(hwnd, 0, reinterpret_cast<LONG>(lpcs->lpCreateParams));
			PostMessage(hwnd, WM_USER+1, 0, 0);
		} break;
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			window_renderer * wr = reinterpret_cast<window_renderer *>(GetWindowLong(hwnd, 0));
			HDC hdc = BeginPaint(hwnd, &ps);
			WaitForSingleObject(wr->_hMutex, INFINITE);
			SetDIBitsToDevice(hdc, 
				0, 0, 
				wr->get_bmi()->bmiHeader.biWidth, abs(wr->get_bmi()->bmiHeader.biHeight), 
				0, 0,
				0, abs(wr->get_bmi()->bmiHeader.biHeight),
				wr->_bitmap, wr->get_bmi(), DIB_RGB_COLORS);
			ReleaseMutex(wr->_hMutex);
			EndPaint(hwnd, &ps);
		} break;
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

DWORD CALLBACK window_renderer::threadProc(void * wnd)
{
	WNDCLASS wc;
	wc.style = 0;
	wc.lpfnWndProc = windowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 4;
	wc.hInstance = NULL;
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = "myclassname";
	RegisterClass(&wc);
	HWND hwnd;
	reinterpret_cast<window_renderer *>(wnd)->_hwnd = hwnd = CreateWindow("myclassname", "SMUSH Viewer", WS_POPUPWINDOW | WS_CAPTION, 
		CW_USEDEFAULT, CW_USEDEFAULT, 
		CW_USEDEFAULT, CW_USEDEFAULT, 
		NULL, NULL, NULL, wnd);
	ShowWindow(hwnd, SW_SHOW);
	UpdateWindow(hwnd);
	MSG msg;
	while(GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return 0;
}

window_renderer::window_renderer() throw(exception) : 
	_bitmap(0), _size(0), _clock(0), _out(0) {
	_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	_hMutex = CreateMutex(NULL, FALSE, NULL);
	_hThread = CreateThread(NULL, 0, threadProc, this, 0, NULL);
	WaitForSingleObject(_hEvent, INFINITE);
}

window_renderer::~window_renderer() throw() {
	if(_out) _out->stop();
	PostMessage(_hwnd, WM_USER, 0, 0);
	WaitForSingleObject(_hThread, INFINITE);
	CloseHandle(_hThread);
	CloseHandle(_hMutex);
	CloseHandle(_hEvent);
	clean();
	if(_out) delete _out;
}

bool window_renderer::init_frame(const point & p) throw(exception) {
	SetWindowPos(_hwnd, NULL, 0, 0, 
		p.get_x() + 2*GetSystemMetrics(SM_CXFIXEDFRAME), 
		p.get_y() + 2*GetSystemMetrics(SM_CYFIXEDFRAME) + GetSystemMetrics(SM_CYCAPTION), 
		SWP_NOMOVE|SWP_NOZORDER);
	return bitmap_renderer::init_frame(p);
}
void window_renderer::dump(char * data, int size) throw(exception) {
	ostringstream oss;
	oss << "SMUSH Viewer : " << get_filename() << " : " << get_frame() + 1 << "/" << get_nbframes();
	SetWindowText(_hwnd, oss.str().c_str());
	WaitForSingleObject(_hMutex, INFINITE);
	if((!_bitmap) || (_size != size)) {
		clean();
		_bitmap = new char[size];
		_size = size;
	}
	if(!_bitmap) {
		throw runtime_error("window_renderer unable to allocate frame buffer");
	}
	memcpy(_bitmap, data, size);
	ReleaseMutex(_hMutex);
	InvalidateRect(_hwnd, NULL, FALSE);
}

void window_renderer::clean() throw() {
	if(_bitmap) {
		delete []_bitmap;
		_bitmap = 0;
	}
}

bool window_renderer::wait(int ms) throw() {
	int old = _clock;
	_clock = GetTickCount();
	if(_out && _out->wait()) return true;
	if(!old) return true;
	old += ms;
	while(_clock < old) {
		//~ if((old - _clock) > 2)
			//~ Sleep(1);
		_clock = GetTickCount();
		if(_out && _out->wait()) return true;
	}
	return true;
}

sound_renderer * window_renderer::get_sound_renderer() throw(exception)
{
	if(! _out) _out = new ds_wave_out;
	return _out;
}

bool win32_save_as_wave(const string & fname, char * data, int size, int freq)
{
	HMMIO       hmmio;              // file handle for open file 
	MMCKINFO    mmckinfoParent;     // parent chunk information 
	MMCKINFO    mmckinfoSubchunk;   // subchunk information structure 
	
	WAVEFORMATEX wf;
	wf.wBitsPerSample = 8;
	wf.nChannels = 1;
	wf.nSamplesPerSec = freq;
	wf.nBlockAlign = wf.nChannels * wf.wBitsPerSample / 8;
	wf.wFormatTag = WAVE_FORMAT_PCM;
	wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign;
	wf.cbSize =0;
	
	hmmio = mmioOpen(const_cast<char*>(fname.c_str()), NULL, MMIO_CREATE | MMIO_WRITE | MMIO_ALLOCBUF);
	mmckinfoParent.fccType = mmioFOURCC('W', 'A', 'V', 'E'); 
	mmioCreateChunk(hmmio, &mmckinfoParent, MMIO_CREATERIFF);
	mmckinfoSubchunk.ckid = mmioFOURCC('f', 'm', 't', ' '); 
	mmioCreateChunk(hmmio, &mmckinfoSubchunk, 0);
	mmioWrite(hmmio, reinterpret_cast<char*>(&wf), sizeof(wf));
	mmioAscend(hmmio, &mmckinfoSubchunk, 0);
	mmckinfoSubchunk.ckid = mmioFOURCC('d', 'a', 't', 'a'); 
	mmioCreateChunk(hmmio, &mmckinfoSubchunk, 0);
	mmioWrite(hmmio, data, size);
	mmioAscend(hmmio, &mmckinfoSubchunk, 0);
	mmioAscend(hmmio, &mmckinfoParent, 0);
	mmioClose(hmmio, 0); 
	
	return true;
}
