/*
Copyright (c) 2002 Xavier Trochu

This software is provided 'as-is', without any express or implied warranty. In no event
will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial
applications, and to alter it and redistribute it freely, subject to the following
restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you
   wrote the original software. If you use this software in a product, an acknowledgment
   in the product documentation would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be misrepresented 
   as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/
#ifndef __WIN32_H_
#define __WIN32_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef DEBUG
# ifndef NO_DEBUG_WIN32
#  define DEBUG_WIN32
# endif
#else
# ifdef DEBUG_WIN32
#  error DEBUG_WIN32 defined without DEBUG
# endif
#endif

#include <exception>
#include <string>
#include <windows.h> // for BITMAPINFO, ...

#include "brenderer.h"

/*!	@brief ::renderer implementation that convert frames to windows DIB.

	This class contains utility functions that convert a frame buffer to a valid windows
	DIB.
*/
class bitmap_renderer : public base_renderer {
private:
	int _padding;			//!< bitmap padding (width modulo 4)
	BITMAPINFO * _bmi;		//!< pointer to the bitmap header and bitmap data.
public:
	static const int HEADER_SIZE = sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD);
protected:
	const BITMAPINFO * get_bmi() const throw() { return _bmi; }
public:
	bitmap_renderer() throw(std::exception);
	virtual ~bitmap_renderer() throw();
	virtual bool init_frame(const point & size) throw(std::exception);
	bool set_palette(const palette & pal) throw();

	virtual void dump(char * data, int size) throw(std::exception) = 0;
	void save(int f=-1) throw(std::exception);
};

/*!	@brief ::renderer implementation that convert saves frames as .bmp files.

*/
class bitmap_file_renderer : public bitmap_renderer {
private:
	std::string _name;			//!< base name of the bitmap sequence to generate.
	BITMAPFILEHEADER _bfi;		//!< the bitmap file header.
public:
	bitmap_file_renderer(const std::string & name) throw();
	virtual ~bitmap_file_renderer() throw() {};
	bool init_frame(const point & size) throw();
	void dump(char * data, int size) throw(std::exception);
	bool wait(int ms) throw() { return true; };
	//~ bool start_decode(int version, int nbframe) throw() { return true; }
	bool save_current() throw();
};

class ds_wave_out;

/*!	@brief ::renderer implementation that convert shows frames in a window.

*/
class window_renderer : public bitmap_renderer {
private:
	HWND _hwnd;			//!< the window handle
	char * _bitmap;			//!< bitmap data of the current frame
	int _size;				//!< size of the bitmap data
	int _clock;				//!< date of the last frame update (in ms)
	HANDLE _hThread;		//!< handle of the thread that handle window messages
	HANDLE _hEvent;			//!< handle to an event use for thread synchronisation
	HANDLE _hMutex;			//!< handle to a mutex to protect write access to _bitmap
	ds_wave_out * _out;		//!< the ::sound_renderer

static LRESULT WINAPI windowProc(HWND, UINT, WPARAM, LPARAM);
static DWORD CALLBACK threadProc(void *);
public:
	window_renderer() throw(std::exception);
	virtual ~window_renderer() throw();
	bool init_frame(const point & size) throw(std::exception);
	void dump(char * data, int size) throw(std::exception);
	bool wait(int ms) throw();
	//~ bool start_decode(int, int) throw() { return true; };
	void clean() throw();
	virtual sound_renderer * get_sound_renderer() throw(std::exception);
};

#endif
